/* ############################ MALP License ############################## */
/* # Fri Jan 18 14:00:00 CET 2013                                         # */
/* # Copyright or (C) or Copr. Commissariat a l'Energie Atomique          # */
/* #                                                                      # */
/* # This software is governed by the CeCILL-C license under French law   # */
/* # and abiding by the rules of distribution of free software.  You can  # */
/* # use, modify and/ or redistribute the software under the terms of     # */
/* # the CeCILL-C license as circulated by CEA, CNRS and INRIA at the     # */
/* # following URL http://www.cecill.info.                                # */
/* #                                                                      # */
/* # The fact that you are presently reading this means that you have     # */
/* # had knowledge of the CeCILL-C license and that you accept its        # */
/* # terms.                                                               # */
/* #                                                                      # */
/* # Authors:                                                             # */
/* #   - BESNARD Jean-Baptiste jean-baptiste.besnard@cea.fr               # */
/* #                                                                      # */
/* ######################################################################## */
#include <svut.h>
#include <iostream>
#include <stdint.h>
#include <omp.h>

using namespace svUnitTest;

#include "Hash_Table.h"


using namespace std;


struct entry
{
	uint64_t id;
	int val;
};


int entry_test(void *pl, uint64_t id, void *arg)
{
	struct entry *e = ( struct entry *)pl;
	
	if( e->id == id )
		return 1;
	
	return 0;
}


struct HT_Cell
{
	struct Hash_Table_Internal ___;
	struct entry ent[10];
};



SVUT_DECLARE_FLAT_TEST(Hash_Table,Init_Release)
{
	struct Hash_Table ht;
	
	Hash_Table_init( &ht, sizeof( struct HT_Cell ), 4142, 10 );

	SVUT_ASSERT_EQUAL( ht.table_size, 4142u );
	SVUT_ASSERT_EQUAL( ht.sizeof_payload, sizeof( struct HT_Cell ) );
	SVUT_ASSERT_EQUAL( ht.coll_avoid, 10 );
	SVUT_ASSERT_EQUAL( ht.intern_payload_size, sizeof( struct entry ) );
	SVUT_ASSERT_NOT_NULL( ht.payload );
	
	Hash_Table_release(  &ht, NULL );

	SVUT_ASSERT_EQUAL( ht.table_size, 0u );
	SVUT_ASSERT_EQUAL( ht.sizeof_payload, 0u );
	SVUT_ASSERT_EQUAL( ht.coll_avoid, 0 );
	SVUT_ASSERT_EQUAL( ht.intern_payload_size, 0u );
	SVUT_ASSERT_NULL( ht.payload );
}

SVUT_DECLARE_FLAT_TEST(Hash_Table,Push_and_get)
{
	struct Hash_Table ht;
	struct entry ent;
	struct entry *pushed_e;
	struct entry *ret_e;

	Hash_Table_init( &ht, sizeof( struct HT_Cell ), 4142, 10 );

	ent.id = 42;
	ent.val = 0x1234;

	pushed_e = (struct entry *)Hash_Table_push( &ht, (void *)&ent,  42 );
	
	SVUT_ASSERT_NOT_NULL( pushed_e );
	
	ret_e = (struct entry *)Hash_Table_get( &ht, 42 , NULL, entry_test);

	SVUT_ASSERT_NOT_NULL( ret_e );
	SVUT_ASSERT_EQUAL( pushed_e, ret_e );
	SVUT_ASSERT_EQUAL( 42u, ret_e->id );
	SVUT_ASSERT_EQUAL( 0x1234, ret_e->val );


	Hash_Table_release(  &ht, NULL );

}

SVUT_DECLARE_FLAT_TEST(Hash_Table,Push_and_get_many)
{
	struct Hash_Table ht;
	uint64_t i;
	struct entry ent;
	struct entry *ret_e;
	
	Hash_Table_init( &ht, sizeof( struct HT_Cell ), 4142, 10 );

	/* Push */
	for( i = 0 ; i < 4096 ; i++ )
	{
		ent.id = i;
		ent.val = i * 0x1234;

		Hash_Table_push( &ht, (void *)&ent,  i );
	}

	/* Retrieve */
	for( i = 0 ; i < 4096 ; i++ )
	{
		ret_e = (struct entry *)Hash_Table_get( &ht, i , NULL, entry_test);

		SVUT_ASSERT_NOT_NULL( ret_e );
		SVUT_ASSERT_EQUAL( i, ret_e->id );
		SVUT_ASSERT_EQUAL( (int)(i * 0x1234), ret_e->val );
	}

	/* Retrieve overflow */
	for( i = 4096 ; i < 16384 ; i++ )
	{
		ret_e = (struct entry *)Hash_Table_get( &ht, i , NULL, entry_test);

		SVUT_ASSERT_NULL( ret_e );
	}


	Hash_Table_release(  &ht, NULL );

}


SVUT_DECLARE_FLAT_TEST(Hash_Table,Push_and_get_many_MT)
{
	struct Hash_Table ht;
	uint64_t i;

	
	Hash_Table_init( &ht, sizeof( struct HT_Cell ), 4142, 10 );

	Hash_Table_use_lock( &ht );

#pragma omp parallel
{	
	struct entry *ret_e;
	/* Push */
	#pragma omp for private(i)
	for( i = 0 ; i < 4096 ; i++ )
	{
		struct entry ent;
		ent.id = i;
		ent.val = i * 0x1234;

		Hash_Table_push( &ht, (void *)&ent,  i );
	}


	/* Retrieve */
	#pragma omp for private(i)
	for( i = 0 ; i < 4096 ; i++ )
	{
		ret_e = (struct entry *)Hash_Table_get( &ht, i , NULL, entry_test);

		SVUT_ASSERT_NOT_NULL( ret_e );
		SVUT_ASSERT_EQUAL( i, ret_e->id );
		SVUT_ASSERT_EQUAL( (int)(i * 0x1234), ret_e->val );
	}


	/* Retrieve overflow */
	#pragma omp for private(i)
	for( i = 4096 ; i < 16384 ; i++ )
	{
		ret_e = (struct entry *)Hash_Table_get( &ht, i , NULL, entry_test);

		SVUT_ASSERT_NULL( ret_e );
	}
}

	Hash_Table_release(  &ht, NULL );

}



SVUT_DECLARE_FLAT_TEST(Hash_Table,Hash_Table_push_unique)
{
	struct Hash_Table ht;
	struct entry ent;
	struct entry *push_e;
	struct entry *ret_e;
	
	Hash_Table_init( &ht, sizeof( struct HT_Cell ), 4142, 10 );

	ent.id = 42;
	ent.val = 0x1234;

	push_e = (struct entry *)Hash_Table_push( &ht, (void *)&ent, 42);

	ent.id = 42;
	ent.val = 0x4321;

	/* Switch an existing element */
	ret_e = (struct entry *)Hash_Table_push_unique( &ht, (void *)&ent , 42, NULL, entry_test );

	SVUT_ASSERT_EQUAL( push_e, ret_e );
	SVUT_ASSERT_EQUAL( ret_e->id , 42u );
	SVUT_ASSERT_EQUAL( ret_e->val , 0x4321 );

	/* Switch on unexisting element */
	ent.id = 128;
	ent.val = 0x5678;

	ret_e = (struct entry *)Hash_Table_push_unique( &ht, (void *)&ent , 128, NULL, entry_test );


	SVUT_ASSERT_NOT_NULL( ret_e );
	SVUT_ASSERT_EQUAL( ret_e->id , 128u );
	SVUT_ASSERT_EQUAL( ret_e->val , 0x5678 );

	Hash_Table_release(  &ht, NULL );

}

SVUT_DECLARE_FLAT_TEST(Hash_Table,Hash_Table_get_an_entry)
{
	struct Hash_Table ht;
	struct entry ent;
	struct entry *pushed_e;
	struct entry *ret_e;

	Hash_Table_init( &ht, sizeof( struct HT_Cell ), 4142, 10 );

	ent.id = 42;
	ent.val = 0x1234;

	pushed_e = (struct entry *)Hash_Table_push( &ht, (void *)&ent,  42 );
	
	SVUT_ASSERT_NOT_NULL( pushed_e );
	
	ret_e = (struct entry *)Hash_Table_get_an_entry( &ht );

	SVUT_ASSERT_NOT_NULL( ret_e );
	SVUT_ASSERT_EQUAL( pushed_e, ret_e );
	SVUT_ASSERT_EQUAL( 42u, ret_e->id );
	SVUT_ASSERT_EQUAL( 0x1234, ret_e->val );


	Hash_Table_release(  &ht, NULL );

}


void Hash_Table_walkthrought_cond_2p( struct Hash_Table *ht, void *second_arg, void (*action)(void *, void *), int (*test_func)(void *, void * ), void *test_value );




void Hash_Table_walkthrought_h(void *pelem)
{
	struct entry *ent = (struct entry *)pelem;
	
	ent->val = 42;
	
	
}

void Hash_Table_walkthrought_2p_h(void *pelem, void *sec_a)
{
	struct entry *ent = (struct entry *)pelem;
	int *sa = (int *)sec_a;
	
	ent->val = 42 + *sa;
	
	
}

void Hash_Table_walkthrought_2p_h(void *pelem, void *sec_a, void *sec_b)
{
	struct entry *ent = (struct entry *)pelem;
	int *sa = (int *)sec_a;
	int *sb = (int *)sec_b;
	
	ent->val = 42 + *sa + *sb;
	
	
}

int Hash_Table_walkthrought_test_f(void *pelem, void *t_value)
{
	struct entry *ent = (struct entry *)pelem;
	uint64_t *v = (uint64_t *)t_value;
	
	if( ent->id == *v )
	{
		return 1;
	}
	
	return 0;
}




SVUT_DECLARE_FLAT_TEST(Hash_Table,Hash_Table_walkthrought)
{
	struct Hash_Table ht;
	uint64_t i;
	struct entry *ret_e;
	
	Hash_Table_init( &ht, sizeof( struct HT_Cell ), 4142, 10 );

	/* Push */

	for( i = 0 ; i < 4096 ; i++ )
	{
		struct entry ent;
		ent.id = i;
		ent.val = i * 0x1234;

		Hash_Table_push( &ht, (void *)&ent,  i );
	}

	int *touch_array = (int *)calloc( 4096, sizeof( int ) );
	
	SVUT_ASSERT_NOT_NULL( touch_array );
	
	
	Hash_Table_walkthrought( &ht, Hash_Table_walkthrought_h );
	
	
	for( i = 0 ; i < 4096 ; i++ )
	{
		ret_e = (struct entry *)Hash_Table_get( &ht, i , NULL, entry_test);

		SVUT_ASSERT_NOT_NULL( ret_e );
		SVUT_ASSERT_EQUAL( ret_e->id, i );
		SVUT_ASSERT_EQUAL( ret_e->val, 42 );
	}
	
	
	free( touch_array );

	Hash_Table_release(  &ht, NULL );

}

SVUT_DECLARE_FLAT_TEST(Hash_Table,Hash_Table_walkthrought_2p)
{
	struct Hash_Table ht;
	uint64_t i;
	struct entry *ret_e;
	
	Hash_Table_init( &ht, sizeof( struct HT_Cell ), 4142, 10 );

	/* Push */

	for( i = 0 ; i < 4096 ; i++ )
	{
		struct entry ent;
		ent.id = i;
		ent.val = i * 0x1234;

		Hash_Table_push( &ht, (void *)&ent,  i );
	}

	int *touch_array = (int *)calloc( 4096, sizeof( int ) );
	
	SVUT_ASSERT_NOT_NULL( touch_array );
	
	int sec_a = 8;

	Hash_Table_walkthrought_2p( &ht, &sec_a, Hash_Table_walkthrought_2p_h );
	
	
	for( i = 0 ; i < 4096 ; i++ )
	{
		ret_e = (struct entry *)Hash_Table_get( &ht, i , NULL, entry_test);

		SVUT_ASSERT_NOT_NULL( ret_e );
		SVUT_ASSERT_EQUAL( ret_e->id, i );
		SVUT_ASSERT_EQUAL( ret_e->val, 42 + sec_a );
	}
	
	
	free( touch_array );

	Hash_Table_release(  &ht, NULL );

}

SVUT_DECLARE_FLAT_TEST(Hash_Table,Hash_Table_walkthrought_3p)
{
	struct Hash_Table ht;
	uint64_t i;
	struct entry *ret_e;
	
	Hash_Table_init( &ht, sizeof( struct HT_Cell ), 4142, 10 );

	/* Push */

	for( i = 0 ; i < 4096 ; i++ )
	{
		struct entry ent;
		ent.id = i;
		ent.val = i * 0x1234;

		Hash_Table_push( &ht, (void *)&ent,  i );
	}

	int *touch_array = (int *)calloc( 4096, sizeof( int ) );
	
	SVUT_ASSERT_NOT_NULL( touch_array );
	
	int sec_a = 8;
	int sec_b = 3;

	Hash_Table_walkthrought_3p( &ht, &sec_a, &sec_b, Hash_Table_walkthrought_2p_h );
	
	
	for( i = 0 ; i < 4096 ; i++ )
	{
		ret_e = (struct entry *)Hash_Table_get( &ht, i , NULL, entry_test);

		SVUT_ASSERT_NOT_NULL( ret_e );
		SVUT_ASSERT_EQUAL( ret_e->id, i );
		SVUT_ASSERT_EQUAL( ret_e->val, 42 + sec_a + sec_b );
	}
	
	
	free( touch_array );

	Hash_Table_release(  &ht, NULL );

}

SVUT_DECLARE_FLAT_TEST(Hash_Table,Hash_Table_walkthrought_cond)
{
	struct Hash_Table ht;
	uint64_t i;
	struct entry *ret_e;
	
	Hash_Table_init( &ht, sizeof( struct HT_Cell ), 4142, 10 );

	/* Push */

	for( i = 0 ; i < 4096 ; i++ )
	{
		struct entry ent;
		ent.id = i;
		ent.val = i * 0x1234;

		Hash_Table_push( &ht, (void *)&ent,  i );
	}

	int *touch_array = (int *)calloc( 4096, sizeof( int ) );
	
	SVUT_ASSERT_NOT_NULL( touch_array );
	
	uint64_t id = 8;

	Hash_Table_walkthrought_cond( &ht, Hash_Table_walkthrought_h, Hash_Table_walkthrought_test_f, &id );
	
	for( i = 0 ; i < 4096 ; i++ )
	{
		ret_e = (struct entry *)Hash_Table_get( &ht, i , NULL, entry_test);

		SVUT_ASSERT_NOT_NULL( ret_e );
		SVUT_ASSERT_EQUAL( ret_e->id, i );
		
		if( i == id )
		{
			SVUT_ASSERT_EQUAL( ret_e->val, 42 );
		}
		else
		{
			SVUT_ASSERT_EQUAL( ret_e->val, (int)(i * 0x1234) );
		}
	}
	
	
	free( touch_array );

	Hash_Table_release(  &ht, NULL );

}

void Hash_Table_walkthrought_cond_2p( struct Hash_Table *ht, void *second_arg, void (*action)(void *, void *), int (*test_func)(void *, void * ), void *test_value );

SVUT_DECLARE_FLAT_TEST(Hash_Table,Hash_Table_walkthrought_cond_2p)
{
	struct Hash_Table ht;
	uint64_t i;
	struct entry *ret_e;
	
	Hash_Table_init( &ht, sizeof( struct HT_Cell ), 4142, 10 );

	/* Push */

	for( i = 0 ; i < 4096 ; i++ )
	{
		struct entry ent;
		ent.id = i;
		ent.val = i * 0x1234;

		Hash_Table_push( &ht, (void *)&ent,  i );
	}

	int *touch_array = (int *)calloc( 4096, sizeof( int ) );
	
	SVUT_ASSERT_NOT_NULL( touch_array );
	
	uint64_t id = 8;
	int sec_a = 8;

	Hash_Table_walkthrought_cond_2p( &ht, &sec_a, Hash_Table_walkthrought_2p_h, Hash_Table_walkthrought_test_f, &id );
	
	for( i = 0 ; i < 4096 ; i++ )
	{
		ret_e = (struct entry *)Hash_Table_get( &ht, i , NULL, entry_test);

		SVUT_ASSERT_NOT_NULL( ret_e );
		SVUT_ASSERT_EQUAL( ret_e->id, i );
		
		if( i == id )
		{
			SVUT_ASSERT_EQUAL( ret_e->val, 42 + sec_a );
		}
		else
		{
			SVUT_ASSERT_EQUAL( ret_e->val, (int)(i * 0x1234) );
		}
	}
	
	
	free( touch_array );

	Hash_Table_release(  &ht, NULL );

}
