/* ############################ MALP License ############################## */
/* # Fri Jan 18 14:00:00 CET 2013                                         # */
/* # Copyright or (C) or Copr. Commissariat a l'Energie Atomique          # */
/* #                                                                      # */
/* # This software is governed by the CeCILL-C license under French law   # */
/* # and abiding by the rules of distribution of free software.  You can  # */
/* # use, modify and/ or redistribute the software under the terms of     # */
/* # the CeCILL-C license as circulated by CEA, CNRS and INRIA at the     # */
/* # following URL http://www.cecill.info.                                # */
/* #                                                                      # */
/* # The fact that you are presently reading this means that you have     # */
/* # had knowledge of the CeCILL-C license and that you accept its        # */
/* # terms.                                                               # */
/* #                                                                      # */
/* # Authors:                                                             # */
/* #   - BESNARD Jean-Baptiste jean-baptiste.besnard@cea.fr               # */
/* #                                                                      # */
/* ######################################################################## */
#include "Analyzer_BB.h"

#include "MALP_Config.h"
#include "Meta_Data_Handler.h"
#include "Trace_Collector.h"
#include "Timer.h"
#include "Debug.h"
#include "Shared_Cache.h"

#include <pthread.h>

struct Analyzer_BB * Analyzer_BB()
{
	return &(Trace_Collector()->anbb);
}

static MALP_Spinlock cockpit_lock = 0;
static volatile uint64_t flying_packs = 0;
static volatile uint64_t processed_packs = 0;

void Analyzer_BB_add_pack()
{
		MALP_Spinlock_lock( &cockpit_lock );
		flying_packs++;
		MALP_Spinlock_unlock( &cockpit_lock );
}

void Analyzer_BB_remove_pack()
{
		MALP_Spinlock_lock( &cockpit_lock );
		flying_packs--;
		processed_packs++;
		MALP_Spinlock_unlock( &cockpit_lock );
}


uint64_t Analyzer_BB_get_pack()
{
	uint64_t ret = 0;
	MALP_Spinlock_lock( &cockpit_lock );
	ret = flying_packs;
	MALP_Spinlock_unlock( &cockpit_lock );
	
	return ret;
}


void Analyzer_BB_KS_UNPACKER( struct MALP_blackboard *bb, struct Data_entry **de, uint32_t entry_count, void *dummy )
{	
	struct Event_Pack * pack = (struct Event_Pack *)Data_entry_get_payload(de[0]);
	struct Analyzer_BB *an = (struct Analyzer_BB *)bb->context;

	int i = 0;

	for( i = 0 ; i  < pack->offset; i++ )
	{
			Event_Dispatcher_call( &an->disp, &pack->events[i] );
	}
	

	Analyzer_BB_remove_pack();
	
   	if( !(processed_packs % 1000) && TA_Conf_get_key_bool(TA_CONF_DISPLAY_BB_LOAD) )
	{

		MALP("Flying : %lld(%g MB) ===  Processed : %lld(%g MB)\n", (long long)flying_packs, ((double)flying_packs * sizeof( struct Event_Pack) ) / (1024*1024),
				(long long)processed_packs,((double)processed_packs * sizeof( struct Event_Pack) ) / (1024*1024));
	}
}

void __Analyzer_BB_push_event(struct MALP_Trace_Event *container, void *pbb);


json_t * Cache_gen_partition_desc( uint64_t partition_id )
{
	json_t *partition_desc = json_object();
	
	VMPI_Partition_desc *desc = VMPI_Get_desc_by_id(partition_id);
	
	if( !desc )
		return partition_desc;
	
	json_t *root_rank = json_int( desc->root_id );
	json_t *partition_size = json_int( desc->size );
	json_t *partition_name = json_string( desc->name );
	json_t *program_name = json_string( desc->program_name );
	json_t *command_line = json_string( desc->command_line );
	json_t *collectors = json_array();
	
	json_object_set( partition_desc, "root_rank", root_rank );
	json_object_set( partition_desc, "partition_size", partition_size );
	json_object_set( partition_desc, "partition_name", partition_name );
	json_object_set( partition_desc, "program_name", program_name );
	json_object_set( partition_desc, "command_line", command_line );
	json_object_set( partition_desc, "collectors", collectors );
	
	json_decref( root_rank );
	json_decref( partition_size );
	json_decref( partition_name );
	json_decref( program_name );
	json_decref( command_line );
	json_decref( collectors );

	return partition_desc;	
}




void Analyzer_BB_init( struct Analyzer_BB *bb , uint64_t max_pack_buff)
{
	/* Set the maximum number of buffered packs */
	bb->max_flying_packs = max_pack_buff;
	
	
	Event_Dispatcher_init( &bb->disp );
	/* Register Meta data handling */
	Event_Dispatcher_register( &bb->disp, MALP_EVENT_META, _Meta_Data_Handler_push , NULL );

	/* Init Blackboard make it serial if we run over OTF2 */
	if( TA_Conf_get_key_bool( TA_CONF_OTF2_ANALYSIS_ENABLED ) )
		MALP_blackboard_init( &bb->bb, 1, (void *)bb );
	else
		MALP_blackboard_init( &bb->bb, 0, (void *)bb );

	/* Register unpacker KS and save dispatcher address */
	uint64_t sens_list[1];
	sens_list[0] = DE_ID("AN_BB", "BB_TYPE_PACK");
	
	/* KS for everything except OTF */
	MALP_Blackboard_new_ks( &bb->bb, KS_ID("AN_BB","BB_KS_UNPACKER"), sens_list, 1, Analyzer_BB_KS_UNPACKER, NULL );
	

	/* Setup report rendering env */

	int partition_count = VMPI_Get_partition_count();
	int i;

	/* Register partitions in the cache */
	int rank;
	PMPI_Comm_rank( VMPI_Get_partition_comm(), &rank );
	
	if( rank == 0 )
	{
		/* Try to get the partition array or create it */
		json_t * partition_array = json_array();
	
		for( i = 0 ; i < partition_count ; i++ )
		{
			/* Register all partitions except the analyzer */
			if( i == VMPI_Get_partition_id() )
				continue;
				
			json_t *desc = Cache_gen_partition_desc( i );
			
			if( desc )
			{
				json_array_push_at(partition_array, i, desc);
				json_decref( desc );
			}	
			
		}
		
		nodeSharedCache_set( &nsc, "partitions", partition_array );
		
		json_decref( partition_array );
	
	}

	
	bb->report = malloc( partition_count * sizeof( struct Report ) );
	
	if( !bb->report )
	{
		perror("malloc");
		abort();
	}
	
	for( i = 0 ; i < partition_count ; i++ )
	{
		if( i != VMPI_Get_partition_id() )
		{
			Report_init( &bb->report[i] );
			Report_populate( &bb->report[i] , i );
			Report_init_coll( &bb->report[i] );
		}
	}
	
	if( rank == 0 )
		nodeSharedCache_commit( &nsc, "partitions" );
}

void Analyzer_BB_release( struct Analyzer_BB *bb )
{

	/* Wait for all packs to be pushed */

	uint64_t pack_count = 1;

	while( pack_count )
	{
		pack_count =  Analyzer_BB_get_pack();
		pthread_yield();
	}

	MALP_blackboard_wait( &bb->bb);
	
	MALP_blackboard_release( &bb->bb );


	/* Reduce duration infos */
	Duration_infos_reduce();

	
	int i;
	int report_count = VMPI_Get_partition_count();
	
	for( i = 0 ; i < report_count ; i++ )
	{
		if( i != VMPI_Get_partition_id() )
		{
			Report_reduce( &bb->report[i] );
			Report_accumulate( &bb->report[i] );
			Report_release( &bb->report[i] );
		}
	}
	
	free( bb->report ) ;
	bb->report = NULL;

	int rank;
	PMPI_Comm_rank( VMPI_Get_partition_comm(), &rank );
	
	if( rank == 0 )
	{
		/* Update all the partition informations (final commit) */
		nodeSharedCache_commit( &nsc, "partitions" );
	}

	Event_Dispatcher_release( &bb->disp );

}


static uint64_t pack_id = 0;

void Analyzer_BB_push_pack( struct Analyzer_BB *bb, struct Event_Pack *pack )
{
	uint64_t pack_count = bb->max_flying_packs + 1;
	uint64_t cnt = 0;
	
	while( bb->max_flying_packs < pack_count )
	{
		pack_count = Analyzer_BB_get_pack();

		/* Let the BB work */
		if( !(cnt % 100 ) )
		{
			pthread_yield();
			cnt = 0;
		}

		cnt++;
	}

	/* Just to avoid some CRCs */
	if(pack_id == 0)
		pack_id = DE_ID("AN_BB", "BB_TYPE_PACK");


	MALP_Blackboard_submit_data( &bb->bb, pack_id, (void *)pack, sizeof( struct Event_Pack ) );

	Analyzer_BB_add_pack();
}



