/* ############################ MALP License ############################## */
/* # Fri Jan 18 14:00:00 CET 2013                                         # */
/* # Copyright or (C) or Copr. Commissariat a l'Energie Atomique          # */
/* #                                                                      # */
/* # This software is governed by the CeCILL-C license under French law   # */
/* # and abiding by the rules of distribution of free software.  You can  # */
/* # use, modify and/ or redistribute the software under the terms of     # */
/* # the CeCILL-C license as circulated by CEA, CNRS and INRIA at the     # */
/* # following URL http://www.cecill.info.                                # */
/* #                                                                      # */
/* # The fact that you are presently reading this means that you have     # */
/* # had knowledge of the CeCILL-C license and that you accept its        # */
/* # terms.                                                               # */
/* #                                                                      # */
/* # Authors:                                                             # */
/* #   - BESNARD Jean-Baptiste jean-baptiste.besnard@cea.fr               # */
/* #                                                                      # */
/* ######################################################################## */
#include "Array.h"

#include <math.h>
#include "Debug.h"

#include "VMPI.h"

/*-----------------------------------------------------------
INIT
-----------------------------------------------------------*/


static inline double * __Array_init_datas( size_t size )
{
	double *tmp = calloc( size, sizeof( double ) );

	if( !tmp ) {
		printf("Failled to allocate datas ( %f MB ) \n", (float)(size * sizeof( double) ) / ( 1024 * 1024 ) );
		return NULL;
	} 
	//else 
	//{
		//printf("Initializing a %f MB Array \n",  (float)(size * sizeof( double) ) / ( 1024 * 1024 ) );
	//}

	return tmp;
}


void Array_init_sparse_data(struct Array *dts, int dimension)
{

	dts->datas = NULL;

	double cell_size = 0;

	switch(dimension)
	{
		case 1 :
			cell_size = sizeof(struct Array_1D_point_cell);
		break;
		case 2 :
			cell_size = sizeof(struct Array_2D_point_cell);
		break;
		case 3 :
			cell_size = sizeof(struct Array_3D_point_cell);
		break;
		case 4 :
			cell_size = sizeof(struct Array_4D_point_cell);
		break;
		default:
			ERROR("No such dimension %d\n", dimension);
			exit(1);

	}

	Red_Ht_init( &dts->rht, cell_size, 16384, Array_COLL_AVOID);

}


void Array_zero_limits( struct Array *dts )
{

	dts->min = 0;
	dts->max = 0;
	dts->moy =  0;
	dts->tmin = 0;
	dts->tmax = 0;

	dts->cmin = 0;
	dts->cmax = 0;
	dts->cmoy =  0;
	dts->ctmin = 0;
	dts->ctmax = 0;

}

void Array_copy_limits( struct Array *dts , struct Array *src )
{

	dts->min = src->min;
	dts->max = src->max;
	dts->moy =  src->moy;
	dts->tmin = src->tmin;
	dts->tmax = src->tmax;

	dts->cmin = src->cmin;
	dts->cmax = src->cmax;
	dts->cmoy =  src->cmoy;
	dts->ctmin = src->ctmin;
	dts->ctmax = src->ctmax;

}



int Array_init_1D (struct Array * dts, int Dx, int is_sparse)
{

	if( Dx < 0 ) {
		printf("Bad dimension %d \n", Dx);
		return 1;
	}

	dts->is_sparse = is_sparse;

	if( !is_sparse )
	{
		dts->datas =  __Array_init_datas( Dx );

		if( !dts->datas ) {
			printf("Failed to allocate \n");
			return 1;
		}
	}
	else
	{
		Array_init_sparse_data(dts, 1);
	}

	dts->dimension = 1;
	dts->Dx = Dx;
	dts->Dy = 0;
	dts->Dz = 0;
	dts->Dt = 0;
	dts->has_datas = 0;

	Array_zero_limits( dts );

	return 0;
}


int Array_init_2D (struct Array * dts, int Dx, int Dy, int is_sparse)
{
	if( (Dx < 0) || (Dy < 0) ) {
		printf("Bad dimension (%d, %d) \n", Dx, Dy);
		return 1;
	}

	dts->is_sparse = is_sparse;

	if( !is_sparse )
	{
		dts->datas =  __Array_init_datas( Dx * Dy );

		if( !dts->datas ) {
			printf("Failed to allocate \n");
			return 1;
		}
	}
	else
	{
		Array_init_sparse_data(dts, 2);
	}

	dts->dimension = 2;
	dts->Dx = Dx;
	dts->Dy = Dy;
	dts->Dz = 0;
	dts->Dt = 0;
	dts->has_datas = 0;

	Array_zero_limits( dts );

	return 0;
}


int Array_init_3D (struct Array * dts, int Dx, int Dy, int Dz, int is_sparse)
{
	if( (Dx < 0) || (Dy < 0) || (Dz < 0) ) {
		printf("Bad dimension (%d,%d,%d) \n", Dx, Dy, Dz);
		return 1;
	}

	dts->is_sparse = is_sparse;

	if( !is_sparse )
	{
		dts->datas =  __Array_init_datas( Dx * Dy * Dz );

		if( !dts->datas ) {
			printf("Failed to allocate \n");
			return 1;
		}
	}
	else
	{
		Array_init_sparse_data(dts, 3);
	}

	dts->dimension = 3;
	dts->Dx = Dx;
	dts->Dy = Dy;
	dts->Dz = Dz;
	dts->Dt = 0;
	dts->has_datas = 0;

	Array_zero_limits( dts );

	return 0;
}


int Array_init_4D (struct Array * dts, int Dx, int Dy, int Dz, int Dt, int is_sparse)
{
	if( (Dx < 0) || (Dy < 0) || (Dz < 0) || (Dt < 0) ) {
		printf("Bad dimension (%d,%d,%d,%d) \n", Dx, Dy, Dz, Dt);
		return 1;
	}

	dts->is_sparse = is_sparse;

	if( !is_sparse )
	{
		dts->datas =  __Array_init_datas( Dx * Dy * Dz * Dt );

		if( !dts->datas ) {
			printf("Failed to allocate \n");
			return 1;
		}
	}
	else
	{
		Array_init_sparse_data(dts, 4);
	}

	dts->dimension = 4;
	dts->Dx = Dx;
	dts->Dy = Dy;
	dts->Dz = Dz;
	dts->Dt = Dt;
	dts->has_datas = 0;

	Array_zero_limits( dts );

	return 0;
}



void Array_copy_datas(  struct Array *dts, struct Array *source )
{
	if( !source->is_sparse )
	{
		/* Direct copy from buffer */
		switch( source->dimension ) {
			case 1:
				memcpy(dts->datas, source->datas, dts->Dx * sizeof(double));
				break;
			case 2 :
				memcpy(dts->datas, source->datas, dts->Dx * dts->Dy * sizeof(double));
				break;
			case 3:
				memcpy(dts->datas, source->datas, dts->Dx * dts->Dy * dts->Dz * sizeof(double));
				break;
			case 4 :
				memcpy(dts->datas, source->datas, dts->Dx * dts->Dy * dts->Dz * dts->Dt * sizeof(double));
				break;
		}
	}
	else
	{
		/* Copy all the data back */
		int x=0;
		int y=0;
		int z=0;
		int t=0;

		switch( source->dimension ) {
			case 4 :
				for( t = 0 ; t < dts->Dt ; t++ ) {
				case 3 :
					for( z = 0 ; z < dts->Dz ; z++ ) {
					case 2 :
						for( y = 0 ; y < dts->Dy ; y++ ) {
						case 1 :
							for( x = 0 ; x < dts->Dx ; x++ ) {
								double value = Array_dim_agnostic_get( source , x, y, z, t );
								Array_dim_agnostic_set( dts , value, x, y, z, t );
							}
						}

					}
				}
		}
	}

	Array_copy_limits( dts , source );
}


int Array_copy( struct Array *dts, struct Array *source )
{
	int ret = 0;

	if( !dts || !source ) {
		printf("cannot copy without Arrays\n");
		return 1;
	}

	//printf("About to cpy : x : %d y : %d  z : %d \n", source->Dx , source->Dy, source->Dz);

	switch( source->dimension ) {
		case 1:
			ret = Array_init_1D(dts, source->Dx , source->is_sparse);
			break;
		case 2 :
			ret = Array_init_2D(dts, source->Dx , source->Dy  , source->is_sparse);
			break;
		case 3:
			ret = Array_init_3D(dts, source->Dx , source->Dy, source->Dz , source->is_sparse);
			break;
		case 4 :
			ret = Array_init_4D(dts, source->Dx , source->Dy, source->Dz, source->Dt , source->is_sparse);
			break;
	}

	if( ret ) {
		printf("Array copy failed \n");
		return ret;
	}

	Array_copy_datas( dts, source );


	return 0;
}



void Array_release (struct Array * dts)
{
	if( dts->datas ) {
		free( dts->datas );
		dts->datas = NULL;
	}

	if(dts->is_sparse)
	{
		Red_Ht_release(&dts->rht, NULL );
	}

	dts->dimension = 0;

	dts->Dx = 0;
	dts->Dy = 0;
	dts->Dz = 0;
	dts->Dt = 0;

	Array_zero_limits( dts );

}


/*-----------------------------------------------------------
MISC
-----------------------------------------------------------*/



void Array_1D_point_reduce(void *pa, void *pb)
{
	struct Array_1D_point *a = (struct Array_1D_point *) pa;
	struct Array_1D_point *b = (struct Array_1D_point *) pb;

	a->value += b->value;
}

int Array_1D_point_test(void *pa, uint64_t key, void *pc)
{
	struct Array_1D_point *a = (struct Array_1D_point *) pa;
	struct Array_1D_point *t = (struct Array_1D_point *)pc;

	if(key != a->key)
		return 0;

#if FORCE_CRC_COLLISION_CHEK_ARRAY
	if(a->x == t->x)
		return 1;
#else
	return 1;
#endif


	return 0;
}

uint64_t Array_1D_point_key(void *pa)
{
	struct Array_1D_point *a = (struct Array_1D_point *) pa;

	return a->key;
}

void Array_2D_point_reduce(void *pa, void *pb)
{
	struct Array_2D_point *a = (struct Array_2D_point *) pa;
	struct Array_2D_point *b = (struct Array_2D_point *) pb;

	a->value += b->value;
}

int Array_2D_point_test(void *pa, uint64_t key, void *pc)
{
	struct Array_2D_point *a = (struct Array_2D_point *) pa;
	struct Array_2D_point *t = (struct Array_2D_point *)pc;

	if(key != a->key)
		return 0;

#if FORCE_CRC_COLLISION_CHEK_ARRAY
	if( (a->x == t->x) &&  (a->y == t->y) )
		return 1;
#else
	return 1;
#endif


	return 0;
}

uint64_t Array_2D_point_key(void *pa)
{
	struct Array_2D_point *a = (struct Array_2D_point *) pa;

	return a->key;
}


void Array_3D_point_reduce(void *pa, void *pb)
{
	struct Array_3D_point *a = (struct Array_3D_point *) pa;
	struct Array_3D_point *b = (struct Array_3D_point *) pb;

	a->value += b->value;
}

int Array_3D_point_test(void *pa, uint64_t key, void *pc)
{
	struct Array_3D_point *a = (struct Array_3D_point *) pa;
	struct Array_3D_point *t = (struct Array_3D_point *)pc;

	if(key != a->key)
		return 0;

#if FORCE_CRC_COLLISION_CHEK_ARRAY
	if( (a->x == t->x) &&  (a->y == t->y) &&  (a->z == t->z) )
		return 1;
#else
	return 1;
#endif


	return 0;
}

uint64_t Array_3D_point_key(void *pa)
{
	struct Array_3D_point *a = (struct Array_3D_point *) pa;

	return a->key;
}



void Array_4D_point_reduce(void *pa, void *pb)
{
	struct Array_4D_point *a = (struct Array_4D_point *) pa;
	struct Array_4D_point *b = (struct Array_4D_point *) pb;

	a->value += b->value;
}

int Array_4D_point_test(void *pa, uint64_t key, void *pc)
{
	struct Array_4D_point *a = (struct Array_4D_point *) pa;
	struct Array_4D_point *t = (struct Array_4D_point *)pc;

	if(key != a->key)
		return 0;

#if FORCE_CRC_COLLISION_CHEK_ARRAY
	if( (a->x == t->x) &&  (a->y == t->y) &&  (a->z == t->z) &&  (a->t == t->t))
		return 1;
#else
	return 1;
#endif


	return 0;
}

uint64_t Array_4D_point_key(void *pa)
{
	struct Array_4D_point *a = (struct Array_4D_point *) pa;

	return a->key;
}


void Array_reduce (struct Array * dts, MPI_Op op)
{
	int rank = 0;
	PMPI_Comm_rank(VMPI_Get_partition_comm(), &rank);

	if( !dts->is_sparse )
	{
		size_t size = dts->Dx;

		if( dts->Dy )
			size *= dts->Dy;
		if( dts->Dz )
			size *= dts->Dz;
		if( dts->Dt )
			size *= dts->Dt;


		double *recv_buff =  NULL;


		if( rank == 0 ) {
			recv_buff = __Array_init_datas( size );
		}


		int ret = PMPI_Reduce( dts->datas , recv_buff, size, MPI_DOUBLE , op, 0, VMPI_Get_partition_comm());

		if( ret != MPI_SUCCESS ) {
			printf("MPI_Reduce failed \n");
			abort();
		}

		double *to_free = dts->datas;
		dts->datas = recv_buff;


		free( to_free );
	}
	else
	{
		void (*red_func)(void * ,void *) = NULL;
		int (*test_func)(void *, uint64_t , void *) = NULL;
		uint64_t (*key_func)(void *) = NULL;

		switch(dts->dimension)
		{
			case 1:
				red_func = Array_1D_point_reduce;
				test_func = Array_1D_point_test;
				key_func = Array_1D_point_key;
			break;
			case 2:
				red_func = Array_2D_point_reduce;
				test_func = Array_2D_point_test;
				key_func = Array_2D_point_key;
			break;
			case 3:
				red_func = Array_3D_point_reduce;
				test_func = Array_3D_point_test;
				key_func = Array_3D_point_key;
			break;
			case 4:
				red_func = Array_4D_point_reduce;
				test_func = Array_4D_point_test;
				key_func = Array_4D_point_key;
			break;
			default :
				ERROR("No such dimension %d\n", dts->dimension);
				exit(1);
		}

		Red_Ht_reduce(&dts->rht, red_func, test_func, key_func, NULL );

	}

    
	Array_zero_limits( dts );
}




double Array_dim_agnostic_get( struct Array *dts , int x, int y, int z, int t )
{

	switch( dts->dimension ) {
		case 1 :
			return Array_get_1D( dts, x);
			break;
		case 2 :
			return Array_get_2D( dts, x, y);
			break;
		case 3 :
			return Array_get_3D( dts, x, y, z);
			break;
		case 4 :
			return Array_get_4D( dts, x, y, z, t);
			break;
		default :
			printf("No such dimension \n");
			abort();
	}
}

void Array_dim_agnostic_set( struct Array *dts , double value, int x, int y, int z, int t )
{

	switch( dts->dimension ) {
		case 1 :
			Array_set_1D( dts, value, x);
			break;
		case 2 :
			Array_set_2D( dts, value, x, y);
			break;
		case 3 :
			Array_set_3D( dts, value, x, y, z);
			break;
		case 4 :
			Array_set_4D( dts, value, x, y, z, t);
			break;
		default :
			printf("No such dimension \n");
			abort();
	}
}




/*-----------------------------------------------------------
MIN/MAX/MOY
-----------------------------------------------------------*/


void Array_reset_bounds( struct Array *dts )
{
	dts->cmax = 0;
	dts->cmin = 0;
	dts->cmoy = 0;
	dts->ctmax = 0;
	dts->ctmin = 0;
}

uint64_t Array_count_not_null_2D( struct Array *dts  )
{
	int x,y;

	
	uint64_t ret = 0;

	for( x = 0 ; x < dts->Dx ; x++ )
	{
		for( y = 0 ; y < dts->Dy ; y++ )
		{
			double val = Array_get_2D(dts, x, y);

			if( val )
				ret++;
		}
	}

	return ret;
}



int Array_get_get_max_offset_1D( struct Array *dts )
{
	double max = 0;
	int off = -1;


	int x = 0;

	for ( x = 0 ; x < dts->Dx ; x++ ) {
		double val = Array_get_1D(dts, x);
		if ( val && (max < val) ) {

			off = x;
			max = val;
		}


	}

	if ( 0 <= off )
		Array_set_1D(dts, 0, off);
	return off;
}


int Array_get_get_max_offset_2D_col( struct Array *dts, int col, int set_zero )
{
	double max = 0;
	int off = -1;


	int x = 0;

	for ( x = 0 ; x < dts->Dx ; x++ ) {
		double val = Array_get_2D(dts, x, col);
		if ( val && (max < val) ) {

			off = x;
			max = val;
		}


	}

	if ( 0 <= off && set_zero )
		Array_set_2D(dts, 0, off, col);
	return off;
}

int Array_get_get_max_offset_2D_line( struct Array *dts, int line , int set_zero )
{
	double max = 0;
	int off = -1;


	int y = 0;

	for ( y = 0 ; y < dts->Dy ; y++ ) {
		double val = Array_get_2D(dts, line, y);
		if ( val && (max < val) ) {
			off = y;
			max = val;
		}


	}


	if ( 0 <= off && set_zero )
		Array_set_2D(dts, 0, line, off);

	return off;
}

void Array_zero_datas( struct Array* dts)
{

	int x=0;
	int y=0;
	int z=0;
	int t=0;

	switch( dts->dimension ) {
		case 4 :
			for( t = 0 ; t < dts->Dt ; t++ ) {
			case 3 :
				for( z = 0 ; z < dts->Dz ; z++ ) {
				case 2 :
					for( y = 0 ; y < dts->Dy ; y++ ) {
					case 1 :
						for( x = 0 ; x < dts->Dx ; x++ ) {
							Array_dim_agnostic_set(dts, 0, x, y, z, t);
						}
					}

				}
			}
	}

}


double Array_get_min_but_not_zero (struct Array * dts)
{
	double min = Array_get_max ( dts);

	if( dts->cmin )
		return dts->min;

	int x=0;
	int y=0;
	int z=0;
	int t=0;

	switch( dts->dimension ) {
		case 4 :
			for( t = 0 ; t < dts->Dt ; t++ ) {
			case 3 :
				for( z = 0 ; z < dts->Dz ; z++ ) {
				case 2 :
					for( y = 0 ; y < dts->Dy ; y++ ) {
					case 1 :
						for( x = 0 ; x < dts->Dx ; x++ ) {
							double value = Array_dim_agnostic_get( dts , x, y, z, t );

							if( (value < min) && (value != 0)) {
								min = value;
							}
						}
					}

				}
			}
	}

	dts->cmin = 1;
	dts->min = min;

	return min;
}


double Array_get_min_aggregate_2d (struct Array * dts)
{
	double min = Array_get_max_aggregate_2d ( dts);;

	if( dts->ctmin )
		return dts->tmin;

	int x=0;
	int y=0;

	for( x = 0 ; x < dts->Dx ; x++ ) {

		for( y = 0 ; (y < dts->Dy) && ( y < x ) ; y++ ) {
			double value = Array_dim_agnostic_get( dts , x, y, 0, 0 ) + Array_dim_agnostic_get( dts , y, x, 0, 0 );

			if( (value < min) && value) {
				min = value;
			}
		}
	}

	dts->ctmin = 1;
	dts->tmin = min;

	return min;
}


double Array_get_max_aggregate_2d (struct Array * dts)
{
	double max = 0;

	if( dts->ctmax )
		return dts->tmax;

	int x=0;
	int y=0;

	for( x = 0 ; x < dts->Dx ; x++ ) {

		for( y = 0 ; (y < dts->Dy) && ( y <= x ) ; y++ ) {
			double value = Array_dim_agnostic_get( dts , x, y, 0, 0 ) + Array_dim_agnostic_get( dts , y, x, 0, 0 );

			if( max < value ) {
				max = value;
			}
		}
	}

	dts->ctmax = 1;
	dts->tmax = max;

	return max;
}





double Array_get_min (struct Array * dts)
{
	double min = -42.0;


	int x=0;
	int y=0;
	int z=0;
	int t=0;

	switch( dts->dimension ) {
		case 4 :
			for( t = 0 ; t < dts->Dt ; t++ ) {
			case 3 :
				for( z = 0 ; z < dts->Dz ; z++ ) {
				case 2 :
					for( y = 0 ; y < dts->Dy ; y++ ) {
					case 1 :
						for( x = 0 ; x < dts->Dx ; x++ ) {
							double value = Array_dim_agnostic_get( dts , x, y, z, t );

							if( (value < min) || (min == -42.0) ) {
								min = value;
							}
						}
					}

				}
			}
	}

	return min;
}

double Array_get_max (struct Array * dts)
{
	double max = -42.0;

	if( dts->cmax )
		return dts->max;

	int x=0;
	int y=0;
	int z=0;
	int t=0;

	switch( dts->dimension ) {
		case 4 :
			for( t = 0 ; t < dts->Dt ; t++ ) {
			case 3 :
				for( z = 0 ; z < dts->Dz ; z++ ) {
				case 2 :
					for( y = 0 ; y < dts->Dy ; y++ ) {
					case 1 :
						for( x = 0 ; x < dts->Dx ; x++ ) {
							double value = Array_dim_agnostic_get( dts , x, y, z, t );

							if( (max < value) || (max == -42.0) ) {
								max = value;
							}
						}
					}

				}
			}
	}

	dts->cmax = 1;
	dts->max = max;

	return max;
}





double Array_get_moy(struct Array * dts)
{
	double sum = 0;
	double hits = 0;

	if( dts->cmoy )
		return dts->moy;

	int x=0;
	int y=0;
	int z=0;
	int t=0;

	switch( dts->dimension ) {
		case 4 :
			for( t = 0 ; t < dts->Dt ; t++ ) {
			case 3 :
				for( z = 0 ; z < dts->Dz ; z++ ) {
				case 2 :
					for( y = 0 ; y < dts->Dy ; y++ ) {
					case 1 :
						for( x = 0 ; x < dts->Dx ; x++ ) {
							double value = Array_dim_agnostic_get( dts , x, y, z, t );

							sum += value;
							hits++;
						}
					}

				}
			}
	}

	dts->cmoy = 1;
	dts->moy = sum/hits;

	return (sum/hits);
}


double Array_get_sum(struct Array * dts)
{
	double sum = 0;

	int x=0;
	int y=0;
	int z=0;
	int t=0;

	switch( dts->dimension ) {
		case 4 :
			for( t = 0 ; t < dts->Dt ; t++ ) {
			case 3 :
				for( z = 0 ; z < dts->Dz ; z++ ) {
				case 2 :
					for( y = 0 ; y < dts->Dy ; y++ ) {
					case 1 :
						for( x = 0 ; x < dts->Dx ; x++ ) {
							double value = Array_dim_agnostic_get( dts , x, y, z, t );
							sum += value;
						}
					}

				}
			}
	}

	return sum;
}



void Array_normalize( struct Array * dts , double target_max )
{

	int x=0;
	int y=0;
	int z=0;
	int t=0;

	double max = Array_get_max( dts );

	if( !max )
		return;

	switch( dts->dimension ) {
		case 4 :
			for( t = 0 ; t < dts->Dt ; t++ ) {
			case 3 :
				for( z = 0 ; z < dts->Dz ; z++ ) {
				case 2 :
					for( y = 0 ; y < dts->Dy ; y++ ) {
					case 1 :
						for( x = 0 ; x < dts->Dx ; x++ ) {
							double value = Array_dim_agnostic_get( dts , x, y, z, t );
							Array_dim_agnostic_set( dts, ( value * target_max ) / max, x, y, z , t);
						}
					}

				}
			}
	}

	Array_reset_bounds( dts );
}



int Array_range_2D( struct Array *dts , int x, int y )
{
	if( x < 0 )
		return 0;

	if( dts->Dx <= x )
		return 0;

	if( y < 0 )
		return 0;

	if( dts->Dy <= y )
		return 0;

	return 1;
}

void Array_Convolve( struct Array *dest, struct Array *dts, int passes, Array_Conv_kernel k )
{


	static const int gauss_h_7x7[7][7] = { // SIGMA = 319
		{0 ,0 ,0 ,0,0 ,0 ,0 },
		{0 ,0 ,0 ,0,0 ,0 ,0 },
		{0 ,0 ,0 ,0,0 ,0 ,0 },
		{10,33,71,91,71,33,10},
		{0 ,0 ,0 ,0,0 ,0 ,0 },
		{0 ,0 ,0 ,0,0 ,0 ,0 },
		{0 ,0 ,0 ,0,0 ,0 ,0 }
	};

	static const int gauss_7x7[7][7] = { // SIGMA = 1115
		{1 ,4 ,7 ,10,7 ,4 ,1 },
		{4 ,12,26,33,26,12,4 },
		{7 ,26,55,71,55,26,7 },
		{10,33,71,91,71,33,10},
		{7 ,26,55,71,55,26,7 },
		{4 ,12,26,33,26,12,4 },
		{1 ,4 ,7 ,10,7 ,4 ,1 }
	};

	static const int gauss_laplacian_7x7[7][7] = {
		{2 ,4 ,5 ,5 ,5 ,4 ,2 },
		{4 ,5 ,3 ,0 ,3 ,5 ,4 },
		{5,3 ,-12,-24,-12,5 ,3 },
		{5,0,-24,-40,-24,0,5},
		{5,3 ,-12,-24,-12,5 ,3 },
		{4 ,5 ,3 ,0 ,3 ,5 ,4 },
		{2 ,4 ,5 ,5 ,5 ,4 ,2 },
	};



	if( dts->dimension != 2 ) {
		printf("Cannot apply gauss on non 2D Arrays\n");
		abort();
	}

	if( passes == 0 ) {
		Array_copy(dest, dts);
		return;
	}



	int p;
	int x,y;
	int xx,yy;
	double impulse = 0;
	double sigma = 0;

	switch( k ) {
		case   Array_GAUSSIAN_KERNEL :
			sigma = 1115;
			break;
		case Array_GAUSSIAN_HORIZONTAL_KERNEL :
			sigma = 319;
			break;
		case Array_LAPLACIAN_OF_GAUSSIAN_KERNEL :
			sigma = 1;
			break;
	}


	struct Array new;
	Array_init_2D(&new, dts->Dx, dts->Dy, dts->is_sparse);

	Array_init_2D(dest, dts->Dx, dts->Dy, dts->is_sparse);


	struct Array *src_dts = dts;

	for( p = 0 ; p < passes ; p++ ) {
		for( x = 0 ; x < src_dts->Dx ; x++ ) {
			for( y = 0 ; y < src_dts->Dy ; y++ ) {
				impulse = 0;

				for( xx = -3 ; xx <= 3 ; xx++ ) {
					for( yy = -3 ; yy <= 3 ; yy++ ) {
						double val = 0;
						if( Array_range_2D( src_dts , x+xx, y+yy ) ) {
							val = Array_get_2D(src_dts, x+xx, y+yy);
						}

						switch( k ) {
							case   Array_GAUSSIAN_KERNEL :
								impulse += val*gauss_7x7[xx + 3][yy + 3];
								break;
							case   Array_GAUSSIAN_HORIZONTAL_KERNEL :
								impulse += val*gauss_h_7x7[xx + 3][yy + 3];
								break;
							case Array_LAPLACIAN_OF_GAUSSIAN_KERNEL :
								impulse += val*gauss_laplacian_7x7[xx + 3][yy + 3];
								break;
						}

					}
				}



				Array_set_2D(&new , abs( floor( impulse/sigma )), x,y);

			}
		}

		Array_copy_datas( dest, &new );

		if( 1 < passes ) {
			src_dts = dest;
			Array_zero_datas(&new);
		}

	}

	Array_zero_limits(dest);

	Array_release( &new );



}

/*-----------------------------------------------------------
TEST
-----------------------------------------------------------*/

void Array_test()
{

	struct Array dts;

	int rank = 0;
	int comm_size = 0;

	PMPI_Comm_rank(VMPI_Get_partition_comm(), &rank);
	PMPI_Comm_size(VMPI_Get_partition_comm(), &comm_size);

	if( !rank )
		printf("Now checking Array... ");

	double i,j,k,l;

	Array_init_1D(&dts, 1024, 0);


	for( i = 0 ; i < 1024 ; i++ ) {
		Array_set_1D( &dts, i, i);
	}

	for( i = 0 ; i < 1024 ; i++ ) {
		if( Array_get_1D( &dts, i) != i )
			printf("Error in 1D @ %ld \n", i);
	}

	Array_reduce (&dts, MPI_SUM);

	if( rank == 0 ) {

		for( i = 0 ; i < 1024 ; i++ ) {
			if( Array_get_1D( &dts, i) != comm_size*i )
				printf("Error in 1D @ %ld \n", i);
		}

	}

	Array_release(&dts);


	Array_init_2D(&dts, 1024, 512, 0);

	for( i = 0 ; i < 1024 ; i++ ) {
		for( j = 0 ; j < 512 ; j++ ) {
			Array_set_2D( &dts, i+j, i, j);
		}
	}

	for( i = 0 ; i < 1024 ; i++ ) {
		for( j = 0 ; j < 512 ; j++ ) {
			if( Array_get_2D( &dts, i, j) != (i + j) )
				printf("Error in 2D @ %ld, %ld \n", i, j);
		}
	}

	Array_reduce (&dts, MPI_SUM);

	if( rank == 0 ) {
		for( i = 0 ; i < 1024 ; i++ ) {
			for( j = 0 ; j < 512 ; j++ ) {
				if( Array_get_2D( &dts, i, j) != comm_size*(i + j) )
					printf("Error in 2D @ (%ld, %ld) got %ld instead of %ld \n", i, j, Array_get_2D( &dts, i, j), comm_size*(i + j) );
			}
		}
	}


	Array_release(&dts);


	Array_init_3D(&dts, 128, 32, 128, 0);

	for( i = 0 ; i < 128 ; i++ ) {
		for( j = 0 ; j < 32 ; j++ ) {
			for( k = 0 ; k < 128 ; k++ ) {
				Array_set_3D( &dts, i+j+k, i, j, k);
			}
		}
	}

	for( i = 0 ; i < 128 ; i++ ) {
		for( j = 0 ; j < 32 ; j++ ) {
			for( k = 0 ; k < 128 ; k++ ) {
				if( Array_get_3D( &dts, i, j, k) != (i + j + k) )
					printf("Error in 3D @ %ld, %ld, %ld \n", i, j, k);
			}
		}
	}

	Array_reduce (&dts, MPI_SUM);

	if( rank == 0 ) {

		for( i = 0 ; i < 128 ; i++ ) {
			for( j = 0 ; j < 32 ; j++ ) {
				for( k = 0 ; k < 128 ; k++ ) {
					if( Array_get_3D( &dts, i, j, k) != comm_size*(i + j + k) )
						printf("Error in 3D @ %ld, %ld, %ld \n", i, j, k);
				}
			}
		}

	}

	Array_release(&dts);

	Array_init_4D(&dts, 256, 64, 8, 16, 0);

	for( i = 0 ; i < 256 ; i++ ) {
		for( j = 0 ; j < 64 ; j++ ) {
			for( k = 0 ; k < 8 ; k++ ) {
				for( l = 0 ; l < 16 ; l++ ) {
					Array_set_4D( &dts, i+j+k+l, i, j, k, l);
				}
			}
		}
	}

	for( i = 0 ; i < 256 ; i++ ) {
		for( j = 0 ; j < 64 ; j++ ) {
			for( k = 0 ; k < 8 ; k++ ) {
				for( l = 0 ; l < 16 ; l++ ) {
					if( Array_get_4D( &dts, i, j, k, l) != (i + j + k +l) )
						printf("Error in 4D @ %ld, %ld, %ld, %ld \n", i, j, k, l);
				}
			}
		}
	}

	Array_reduce (&dts, MPI_SUM);

	if( rank == 0 ) {

		for( i = 0 ; i < 256 ; i++ ) {
			for( j = 0 ; j < 64 ; j++ ) {
				for( k = 0 ; k < 8 ; k++ ) {
					for( l = 0 ; l < 16 ; l++ ) {
						if( Array_get_4D( &dts, i, j, k, l) != comm_size*(i + j + k +l) )
							printf("Error in 4D @ %ld, %ld, %ld, %ld \n", i, j, k, l);
					}
				}
			}
		}

	}

	Array_release(&dts);

}

