#include "Flat_Matrix.h"
#include "MPI_Events.h"

#include "Meta_Data_Handler.h"
#include "Shared_Cache.h"
#include "Timer.h"
#include <math.h>

/*
 * This source code is a mashup of Time_Matrix therefore
 * if you encounter any bugs they might be there too... 
 */

int  Flat_Matrix_init_coll( struct Collector * coll )
{
	coll->internal_data = malloc( sizeof( struct Flat_Matrix ) );
	
	if(! coll->internal_data )
	{
		perror("malloc");
		abort();
	}
	
	struct Flat_Matrix *tm = (struct Flat_Matrix *)coll->internal_data;
	
	
	int i,j;
	
	for( i = 0 ; i < T_MPI_COUNT ; i++ )
		for( j = 0 ; j < FLAT_GRID_COUNT ; j++ )
			Sample_Grid_init( &tm->mpi_grids[i][j] , 2048 , 1);

	for( i = 0 ; i < T_POSIX_COUNT ; i++ )
		for( j = 0 ; j < FLAT_GRID_COUNT ; j++ )
			Sample_Grid_init( &tm->posix_grids[i][j] , 2048 , 1);	
	
	int rank;
	PMPI_Comm_rank( VMPI_Get_partition_comm(), &rank );

	return 0;
}


int  Flat_Matrix_release_coll( struct Collector * coll )
{
	struct Flat_Matrix *tm = (struct Flat_Matrix *)coll->internal_data;
	
	int i,j;
	
	for( i = 0 ; i < T_MPI_COUNT ; i++ )
		for( j = 0 ; j < FLAT_GRID_COUNT ; j++ )
			Sample_Grid_release( &tm->mpi_grids[i][j] );

	for( i = 0 ; i < T_POSIX_COUNT ; i++ )
		for( j = 0 ; j < FLAT_GRID_COUNT ; j++ )
			Sample_Grid_release( &tm->posix_grids[i][j] );
	
	free( coll->internal_data );

	return 0;
}


void Flat_Matrix_push_event(struct MALP_Trace_Event *event, void *pcoll)
{
	struct Collector *coll = (struct Collector *)pcoll;
	struct Flat_Matrix *tm = (struct Flat_Matrix *)coll->internal_data;


	int type = -1;
	
	if( event->type == MALP_EVENT_MPI )
	{
		struct MPI_Event *mpievt = &event->event.mpi_event;
		type = mpievt->type;

		/* Update the global profile */
		
		Sample_Grid_accumulate( &tm->mpi_grids[type][FLAT_GRID_HITS], 0, event->timestamp / Timer_second(), 1 );
		Sample_Grid_accumulate( &tm->mpi_grids[type][FLAT_GRID_TIME], 0, event->timestamp / Timer_second(), event->duration / Timer_second() );
		
		if( mpievt->size )
			Sample_Grid_accumulate( &tm->mpi_grids[type][FLAT_GRID_SIZE], 0, event->timestamp / Timer_second(), mpievt->size );

	}
	else
	{
		if( event->type == MALP_EVENT_WRAPPED )
		{
			struct Wrapped_Event *we = &event->event.wrapped_event;

			type = we->type;
			
			//printf("POSIX : %s DUR : %g\n", Wrapped_symbol_names[ type ],  event->duration / Timer_second() );
			
			Sample_Grid_accumulate( &tm->posix_grids[type][FLAT_GRID_HITS], 0, event->timestamp / Timer_second(), 1 );
			Sample_Grid_accumulate( &tm->posix_grids[type][FLAT_GRID_TIME], 0, event->timestamp / Timer_second(), event->duration / Timer_second() );
		
			if( we->size )
				Sample_Grid_accumulate( &tm->posix_grids[type][FLAT_GRID_SIZE], 0, event->timestamp / Timer_second(), we->size );
			
		
		}
	}


}


void Flat_Matrix_perform_reduce( struct Collector * coll , struct Sample_Grid *grid )
{
	int j;
	int partition_size = Collector_partition_size( coll );
	
	for( j = 0 ; j < FLAT_GRID_COUNT ; j++ )
	{
			if( grid[j].has_data )
			{
				/* Do a '0' hit over the matrix in order to make all matrix having the same quantum */
				double end_time = Duration_infos_get_end( coll->partition_id )/ Timer_second(); 
				Sample_Grid_scale( &grid[j], end_time*1.05 );
				
				/* We only flatten temporal grids */
				if( j == FLAT_GRID_TIME )
				{
					Sample_Grid_flatten( &grid[j], partition_size );
				}
			}
			
			Sample_Grid_reduce( &grid[j] );
	}
	
}


int  Flat_Matrix_reduce( struct Collector * coll )
{
	struct Flat_Matrix *tm = (struct Flat_Matrix *)coll->internal_data;

	int i;
	
	for( i = 0 ; i < T_MPI_COUNT ; i++ )
	{
		Flat_Matrix_perform_reduce( coll, tm->mpi_grids[i] );
	}
	
	for( i = 0 ; i < T_POSIX_COUNT ; i++ )
	{
		Flat_Matrix_perform_reduce( coll, tm->posix_grids[i] );
	}


	return 0;
}

json_t * Flat_Matrix_gen_entry( struct Sample_Grid * grids, char *name )
{
	json_t * ret = json_object();
	
	json_t * jname = json_string( name );
	json_object_set( ret, "name", jname );
	json_decref( jname );
	
	
	json_t * measures = json_object();
	
	int i;
	
	for( i = 0 ; i < FLAT_GRID_COUNT ; i++ )
	{
		if( !grids[i].has_data )
		{
			continue;
		}
		
		json_t * tmp = Sample_Grid_dump_JSON( &grids[i] );
		
		json_object_set( measures, (char *)flat_grid_types_names[i], tmp );
		json_decref( tmp );
	}
	
	
	json_object_set( ret, "measures", measures );
	
	json_decref( measures );
	
	return ret;
}



int  Flat_Matrix_accumulate( struct Collector * coll )
{
	struct Flat_Matrix *tm = (struct Flat_Matrix *)coll->internal_data;

	int rank;
	PMPI_Comm_rank( VMPI_Get_partition_comm(), &rank );

	if( rank )
		return 0;

	/* Only process 0 from here */

	json_t * flat_matrices = json_object();
	
	json_t * measures = json_array();
	
	
	
	int i;
	
	for( i = 0 ; i < T_MPI_COUNT ; i++ )
	{
		if( tm->mpi_grids[i][0].has_data )
		{
		
			json_t * ret = Flat_Matrix_gen_entry( tm->mpi_grids[i], (char *)MPI_Event_names[ i ]  );
			
			json_array_push( measures, ret );
			
			json_decref( ret );
		
		}
		
	}

	for( i = 0 ; i < T_POSIX_COUNT ; i++ )
	{
		if( tm->posix_grids[i][0].has_data )
		{
		
			json_t * ret = Flat_Matrix_gen_entry( tm->posix_grids[i], (char *)Wrapped_symbol_names[ i ]  );
			
			json_array_push( measures, ret );
			
			json_decref( ret );
		
		}
		
	}
	
	json_object_set( flat_matrices, "measures", measures );
	json_decref( measures );
	
	
	char uri[100];
	snprintf(uri, 100, "%s", coll->analysis_cache_name );
	nodeSharedCache_set( &nsc, uri, flat_matrices );
	
	json_decref( flat_matrices );


	return 0;
}




void Flat_Matrix_init(struct Collector *coll, uint64_t partition_id)
{
		
	Collector_init( coll, MALP_EVENT_ANY,
					partition_id,
					"Flat_Matrix",
					Flat_Matrix_init_coll,
					Flat_Matrix_release_coll,
					Flat_Matrix_push_event,
					Flat_Matrix_reduce,
					Flat_Matrix_accumulate );
	
}
