#include "MALP_OTF2.h"

#include "MPI_Events.h"
#include "Meta_Data_Handler.h"
#include "Shared_Cache.h"
#include "Timer.h"
#include "OTF_Event_forge.h"

#include <sys/time.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>


//*****************************************//
//*            Static data                *//
//*****************************************//
/**
 * @ingroup OTF2_Internal_
 * @brief Timestamp used to know when it is finished (updated at every event)
 */
static OTF2_TimeStamp postFlushTimeStamp;

/**
 * @ingroup OTF2_Internal_
 * @brief The OTF2 archive
 */
static OTF2_Archive* archive = NULL;

/**
 * @ingroup OTF2_Internal_
 * @brief Used to store communicators (uint64_t)
 */
static struct Chained_List communicators;

/**
 * @ingroup OTF2_Internal_
 * @brief Used to store locations (MALP_OTF2_Location)
 */
static struct Hash_Table locations;

/**
 * @ingroup OTF2_Internal_
 * @brief Used to store the ranks (MALP_OTF2_LocationGroup)
 */
static struct Hash_Table ranks;

/**
 * @ingroup OTF2_Internal_
 * @brief Used to store the nodes (MALP_OTF2_LocationGroup)
 */
static struct Hash_Table nodes;


//*****************************************//
//*           OTF2 callbacks              *//
//*****************************************//
OTF2_FlushType pre_flush(   void* userData,
                            OTF2_FileType fileType,
                            OTF2_LocationRef location,
                            void* callerData,
                            bool final )
{
    return OTF2_FLUSH;
}

OTF2_TimeStamp post_flush(  void* userData,
                            OTF2_FileType fileType,
                            OTF2_LocationRef location )
{
    return postFlushTimeStamp;
}

//Flush callbacks

static struct OTF2_FlushCallbacks flush_callbacks = {   .otf2_pre_flush = pre_flush,   .otf2_post_flush = post_flush };


//***************************************************//
//*           OTF2 Location management              *//
//***************************************************//

int MALP_OTF2_LocationGroup_compare(void* elem, uint64_t id, void* extra)
{
    struct MALP_OTF2_LocationGroup* loc = (struct MALP_OTF2_LocationGroup*)elem;
    
    if(loc->id == id)
        return 1;
    
    return 0;
}

int MALP_OTF2_Location_compare(void* elem, uint64_t id, void* extra)
{
    struct MALP_OTF2_Location* loc = (struct MALP_OTF2_Location*)elem;
    
    if(loc->id == id)
        return 1;
    
    return 0;
}

void MALP_OTF2_Location_init(struct MALP_OTF2_Location* loc, struct MALP_Trace_Event *evt)
{
    loc->id = 0;
    loc->last_timestamp = 0;
    loc->event_count = 0;
    loc->evt_writer = NULL;
    loc->def_writer = NULL;

    /* Initialize location from Metada event */
    if( evt != NULL )
    {
		/* Sanity check */
		if( evt->type != MALP_EVENT_META )
		{
			printf("Wrong event type in %s\n", __FUNCTION__ );
			abort();
		}
		
		if( evt->event.stream_descriptor.desc_type != DESC_BEGIN )
		{
			printf("You can only initialize a location with a DESC_BEGIN meta event in %s\n", __FUNCTION__ );
			abort();
		}
		
		/* Stream */
        memcpy( &loc->stream, &evt->event.stream_descriptor, sizeof( struct Stream_Descriptor) );
        /* id */
        loc->id = loc->stream.tid;
        /* rank */
        loc->rank_id = loc->stream.mpi_rank;
        /* Host name/node id */
        loc->node_id = MALP_Trace_crc64(loc->stream.hostname, strlen(loc->stream.hostname));
        /* Writers */
        loc->evt_writer = OTF2_Archive_GetEvtWriter( archive, loc->id );
        loc->def_writer = OTF2_Archive_GetDefWriter( archive, loc->id );
	}
}

void MALP_OTF2_Location_release(void* elem)
{
    struct MALP_OTF2_Location* loc = (struct MALP_OTF2_Location*)elem;

    loc->id = 0;
    loc->last_timestamp = 0;
    loc->event_count = 0;
    loc->evt_writer = NULL;
    loc->def_writer = NULL;

}

void MALP_OTF2_Location_print(void* elem)
{
    struct MALP_OTF2_Location* loc = (struct MALP_OTF2_Location*)elem;
    if(elem != NULL)
    {
#if 0
        printf("id %lu\t", loc->id);
        printf("ts %lu\t", loc->last_timestamp);
        printf("ec %lu\t", loc->event_count);
        printf("node % lu\t", loc->node_id);
        printf("rank %lu\t", loc->rank_id);
        printf("ew %p\t", loc->evt_writer);
        printf("dw %p\t", loc->def_writer);
        printf("\n");
#else
        printf("Location %lu {\n", loc->id);
        printf("    rank %lu\n", loc->rank_id);
        printf("    node %s (%lu)\n", loc->stream.hostname, loc->node_id);
        printf("    omp thread %lu\n", loc->stream.omp_thread);
        printf("}\n");
        
#endif
    } else {
        printf("NULL\n");
    }
}


void MALP_OTF2_Location_send_to_rank(void* elem, int rank)
{
    PMPI_Send(elem, sizeof(struct MALP_OTF2_Location), MPI_CHAR, rank, 4242, VMPI_Get_partition_comm());
}

void MALP_OTF2_Location_recv_from_rank(int rank)
{
    struct MALP_OTF2_Location ldesc;
    PMPI_Recv(&ldesc, sizeof(struct MALP_OTF2_Location), MPI_CHAR, rank, 4242, VMPI_Get_partition_comm(), NULL);
    
    //Set the writers to NULL as they are not needed in other ranks
    //And if they are initialized in other ranks, the resulting trace file may not contain all locations...
    
    ldesc.evt_writer = NULL;
    ldesc.def_writer = NULL;
    
    Hash_Table_push_unique(&locations, &ldesc, ldesc.id, NULL, MALP_OTF2_Location_compare);
}

void MALP_OTF2_Location_send_to_root(void* elem)
{
    MALP_OTF2_Location_send_to_rank(elem, 0);
}

struct MALP_OTF2_Location* MALP_OTF2_Location_get_by_id(uint64_t tid)
{
    struct MALP_OTF2_Location* ldesc = Hash_Table_get(&locations, tid, NULL, MALP_OTF2_Location_compare );
    if( !ldesc )
    {
        printf("Invalid request for location desc (wanted id : %lu)\n", tid);
        return NULL;
    }
    return ldesc;
}

void MALP_OTF2_Location_find_by_rank(void* elem, void* second, void* third)
{
    struct MALP_OTF2_Location* ldesc = (struct MALP_OTF2_Location*)elem;
    uint64_t* rank = (uint64_t*)second;
    struct MALP_OTF2_Location** dest = (struct MALP_OTF2_Location**)third;
    if(ldesc->stream.mpi_rank == (*rank))
    {
        *dest = ldesc;
    }
}

struct MALP_OTF2_Location* MALP_OTF2_Location_get_by_rank(uint64_t rank)
{
    struct MALP_OTF2_Location* ldesc = NULL;
    Hash_Table_walkthrought_3p(&locations, &rank, &ldesc, MALP_OTF2_Location_find_by_rank);
    if( !ldesc )
    {
        printf("Invalid request for location desc (rank : %lu)\n", rank);
        abort();
    }
    return ldesc;
}

//************************************************************//
//          OTF2 Parallel callbacks implementation            //
//************************************************************//

struct OTF2_CollectiveContext
{
	MPI_Comm comm;
};

OTF2_CollectiveCallbacks coll_cb;
OTF2_CollectiveContext GlobalCOMM;
OTF2_CollectiveContext LocalCOMM;

OTF2_CallbackCode
otf2_collectives_get_size( OTF2_Archive*           archive,
                           OTF2_CollectiveContext* commContext,
                           uint32_t*               size )
{
	PMPI_Comm_size( commContext->comm, size );
	
	return OTF2_CALLBACK_SUCCESS;
}



OTF2_CallbackCode
otf2_collectives_get_rank( OTF2_Archive*           archive,
                           OTF2_CollectiveContext* commContext,
                           uint32_t*               rank )
{
	PMPI_Comm_rank( commContext->comm, rank );
	
	return OTF2_CALLBACK_SUCCESS;
}


OTF2_CallbackCode
otf2_collectives_create_local_comm( OTF2_Archive*            archive,
                                    OTF2_CollectiveContext** localCommContext,
                                    OTF2_CollectiveContext*  globalCommContext,
                                    uint32_t                 globalRank,
                                    uint32_t                 globalSize,
                                    uint32_t                 localRank,
                                    uint32_t                 localSize,
                                    uint32_t                 fileNumber,
                                    uint32_t                 numberOfFiles )
{
	


	return OTF2_CALLBACK_SUCCESS;
}

OTF2_CallbackCode
otf2_collectives_free_local_comm( OTF2_Archive*           archive,
                                  OTF2_CollectiveContext* localCommContext )
{
	
	return OTF2_CALLBACK_SUCCESS;
}


OTF2_CallbackCode
otf2_collectives_barrier( OTF2_Archive*           archive,
                          OTF2_CollectiveContext* commContext )
{
	
	PMPI_Barrier( commContext->comm );
	
	return OTF2_CALLBACK_SUCCESS;
}



OTF2_CallbackCode
otf2_collectives_bcast( OTF2_Archive*           archive,
                        OTF2_CollectiveContext* commContext,
                        void*                   data,
                        uint32_t                numberElements,
                        OTF2_Type               type,
                        uint32_t                root )
{
	switch( type )
	{
		case OTF2_TYPE_UINT8:
			PMPI_Bcast(data, numberElements, MPI_UINT8_T, root, commContext->comm);
		break;
		case OTF2_TYPE_UINT16:
			PMPI_Bcast(data, numberElements, MPI_UINT16_T, root, commContext->comm);
		break;
		case OTF2_TYPE_UINT32:
			PMPI_Bcast(data, numberElements, MPI_UINT32_T, root, commContext->comm);
		break;
		case OTF2_TYPE_UINT64:
			PMPI_Bcast(data, numberElements, MPI_UINT64_T, root, commContext->comm);
		break;
		case OTF2_TYPE_INT8:
			PMPI_Bcast(data, numberElements, MPI_INT8_T, root, commContext->comm);
		break;
		case OTF2_TYPE_INT16:
			PMPI_Bcast(data, numberElements, MPI_INT16_T, root, commContext->comm);
		break;
		case OTF2_TYPE_INT32:
			PMPI_Bcast(data, numberElements, MPI_INT32_T, root, commContext->comm);
		break;
		case OTF2_TYPE_INT64:
			PMPI_Bcast(data, numberElements, MPI_INT64_T, root, commContext->comm);
		break;
		case OTF2_TYPE_FLOAT:
			PMPI_Bcast(data, numberElements, MPI_FLOAT, root, commContext->comm);
		break;
		case OTF2_TYPE_DOUBLE:
			PMPI_Bcast(data, numberElements, MPI_DOUBLE, root, commContext->comm);
		break;
		default:
			return OTF2_CALLBACK_INTERRUPT;
	}
	
	return OTF2_CALLBACK_SUCCESS;
}



OTF2_CallbackCode
otf2_collectives_gather( OTF2_Archive*           archive,
                         OTF2_CollectiveContext* commContext,
                         void*                   inData,
                         void*                   outData,
                         uint32_t                numberElements,
                         OTF2_Type               type,
                         uint32_t                root )
{
	switch( type )
	{
		case OTF2_TYPE_UINT8:
			PMPI_Gather(inData, numberElements, MPI_UINT8_T, outData,
			           numberElements, MPI_UINT8_T, root, commContext->comm);
		break;
		case OTF2_TYPE_UINT16:
			PMPI_Gather(inData, numberElements, MPI_UINT16_T, outData,
			           numberElements, MPI_UINT16_T, root, commContext->comm);
		break;
		case OTF2_TYPE_UINT32:
			PMPI_Gather(inData, numberElements, MPI_UINT32_T, outData,
			           numberElements, MPI_UINT32_T, root, commContext->comm);
		break;
		case OTF2_TYPE_UINT64:
			PMPI_Gather(inData, numberElements, MPI_UINT64_T, outData,
			           numberElements, MPI_UINT64_T, root, commContext->comm);
		break;
		case OTF2_TYPE_INT8:
			PMPI_Gather(inData, numberElements, MPI_INT8_T, outData,
			           numberElements, MPI_INT8_T, root, commContext->comm);
		break;
		case OTF2_TYPE_INT16:
			PMPI_Gather(inData, numberElements, MPI_INT16_T, outData,
			           numberElements, MPI_INT16_T, root, commContext->comm);
		break;
		case OTF2_TYPE_INT32:
			PMPI_Gather(inData, numberElements, MPI_INT32_T, outData,
			           numberElements, MPI_INT32_T, root, commContext->comm);
		break;
		case OTF2_TYPE_INT64:
			PMPI_Gather(inData, numberElements, MPI_INT64_T, outData,
			           numberElements, MPI_INT64_T, root, commContext->comm);
		break;
		case OTF2_TYPE_FLOAT:
			PMPI_Gather(inData, numberElements, MPI_FLOAT, outData,
			           numberElements, MPI_FLOAT, root, commContext->comm);
		break;
		case OTF2_TYPE_DOUBLE:
			PMPI_Gather(inData, numberElements, MPI_DOUBLE, outData,
			           numberElements, MPI_DOUBLE, root, commContext->comm);
		break;
		default:
			return OTF2_CALLBACK_INTERRUPT;
	}
	
	return OTF2_CALLBACK_SUCCESS;
}



OTF2_CallbackCode
otf2_collectives_gatherv( OTF2_Archive*           archive,
                          OTF2_CollectiveContext* commContext,
                          void*                   inData,
                          uint32_t                inElements,
                          void*                   outData,
                          const uint32_t*         outElements,
                          OTF2_Type               type,
                          uint32_t                root )
{
	
	return OTF2_CALLBACK_SUCCESS;
}



OTF2_CallbackCode
otf2_collectives_scatter( OTF2_Archive*           archive,
                          OTF2_CollectiveContext* commContext,
                          void*                   inData,
                          void*                   outData,
                          uint32_t                numberElements,
                          OTF2_Type               type,
                          uint32_t                root )
{
	switch( type )
	{
		case OTF2_TYPE_UINT8:
			PMPI_Scatter(inData, numberElements, MPI_UINT8_T, outData,
			           numberElements, MPI_UINT8_T, root, commContext->comm);
		break;
		case OTF2_TYPE_UINT16:
			PMPI_Scatter(inData, numberElements, MPI_UINT16_T, outData,
			           numberElements, MPI_UINT16_T, root, commContext->comm);
		break;
		case OTF2_TYPE_UINT32:
			PMPI_Scatter(inData, numberElements, MPI_UINT32_T, outData,
			           numberElements, MPI_UINT32_T, root, commContext->comm);
		break;
		case OTF2_TYPE_UINT64:
			PMPI_Scatter(inData, numberElements, MPI_UINT64_T, outData,
			           numberElements, MPI_UINT64_T, root, commContext->comm);
		break;
		case OTF2_TYPE_INT8:
			PMPI_Scatter(inData, numberElements, MPI_INT8_T, outData,
			           numberElements, MPI_INT8_T, root, commContext->comm);
		break;
		case OTF2_TYPE_INT16:
			PMPI_Scatter(inData, numberElements, MPI_INT16_T, outData,
			           numberElements, MPI_INT16_T, root, commContext->comm);
		break;
		case OTF2_TYPE_INT32:
			PMPI_Scatter(inData, numberElements, MPI_INT32_T, outData,
			           numberElements, MPI_INT32_T, root, commContext->comm);
		break;
		case OTF2_TYPE_INT64:
			PMPI_Scatter(inData, numberElements, MPI_INT64_T, outData,
			           numberElements, MPI_INT64_T, root, commContext->comm);
		break;
		case OTF2_TYPE_FLOAT:
			PMPI_Scatter(inData, numberElements, MPI_FLOAT, outData,
			           numberElements, MPI_FLOAT, root, commContext->comm);
		break;
		case OTF2_TYPE_DOUBLE:
			PMPI_Scatter(inData, numberElements, MPI_DOUBLE, outData,
			           numberElements, MPI_DOUBLE, root, commContext->comm);
		break;
		default:
			return OTF2_CALLBACK_INTERRUPT;
	}
	
	return OTF2_CALLBACK_SUCCESS;
}



OTF2_CallbackCode
otf2_collectives_scatterv( OTF2_Archive*           archive,
                           OTF2_CollectiveContext* commContext,
                           void*                   inData,
                           const uint32_t*         inElements,
                           void*                   outData,
                           uint32_t                outElements,
                           OTF2_Type               type,
                           uint32_t                root )
{
	return OTF2_CALLBACK_SUCCESS;
}

void otf2_collectives_release ( void*                   userData,
                               OTF2_CollectiveContext* globalCommContext,
                               OTF2_CollectiveContext* localCommContext )
{

	
}

//************************************************************//
//          Collector - Analysis component functions          //
//************************************************************//

int  OTF2_init_coll( struct Collector * coll )
{    
	/* Called before processing any event */
    
    int i;
    struct timeval tv;
  
    int rank, size;
    PMPI_Comm_rank( VMPI_Get_partition_comm(), &rank );
    PMPI_Comm_size( VMPI_Get_partition_comm(), &size );
    
    char *archivePath =  TA_Conf_get_key_string(TA_CONF_OTF2_TRACE_PATH);
    char *archiveName =  TA_Conf_get_key_string(TA_CONF_OTF2_TRACE_NAME);

	char full_archive_name[200];
	char tmp[80];


	if( !rank )
	{

		gettimeofday(&tv, NULL);
		strftime(tmp,50,"%Y%m%d_%H%M%S",localtime(&tv.tv_sec));
		snprintf( full_archive_name, 200, "%s_%s", archiveName, tmp );
	}
    
    if( 1 < size )
    {
		/* Here 0 sends its time dependent archive name to all 
		 * as what if the clocks are different .... */
		 PMPI_Bcast(full_archive_name, 200, MPI_CHAR, 0, VMPI_Get_partition_comm());
	}

	/* Open OTF2 archive */
	archive = OTF2_Archive_Open( archivePath, 
							 full_archive_name,
							 OTF2_FILEMODE_WRITE, 
							 OTF2_CHUNK_SIZE_EVENTS_DEFAULT, 
							 OTF2_CHUNK_SIZE_DEFINITIONS_DEFAULT, 
							 OTF2_SUBSTRATE_POSIX,
							 OTF2_COMPRESSION_NONE );


	/* We have to register parallel callbacks in order to
	 * support parallel archives */
	coll_cb.otf2_get_size = (OTF2_Collectives_GetSize)otf2_collectives_get_size;
	coll_cb.otf2_get_rank = (OTF2_Collectives_GetRank)otf2_collectives_get_rank;
	coll_cb.otf2_create_local_comm = (OTF2_Collectives_CreateLocalComm)otf2_collectives_create_local_comm;
	coll_cb.otf2_free_local_comm = (OTF2_Collectives_FreeLocalComm)otf2_collectives_free_local_comm;
	coll_cb.otf2_barrier = (OTF2_Collectives_Barrier)otf2_collectives_barrier;
	coll_cb.otf2_bcast = (OTF2_Collectives_Bcast)otf2_collectives_bcast;
	coll_cb.otf2_gather = (OTF2_Collectives_Gather)otf2_collectives_gather;
	coll_cb.otf2_gatherv = (OTF2_Collectives_Gatherv)otf2_collectives_gatherv;
	coll_cb.otf2_scatter = (OTF2_Collectives_Scatter)otf2_collectives_scatter;
	coll_cb.otf2_scatterv = (OTF2_Collectives_Scatterv)otf2_collectives_scatterv;
	coll_cb.otf2_release = (OTF2_Collectives_Release)otf2_collectives_release;

	GlobalCOMM.comm = VMPI_Get_partition_comm();
	LocalCOMM.comm = VMPI_Get_partition_comm();

	/* Initialize MALP otf metadata */
	OTF_Event_forge_init();

	OTF2_Archive_SetCollectiveCallbacks( archive,
										 &coll_cb,
										NULL,
										&GlobalCOMM,
										&LocalCOMM );

	/* We set the archive creator name */
	if( !rank )
	{
		OTF2_Archive_SetCreator( archive,  "MALP" );
	}
    
    /* Set the flush callbacks. */
    OTF2_Archive_SetFlushCallbacks( archive, &flush_callbacks, NULL );
    
    /* Initalize communicator list */
    Chained_List_init( &communicators, 32, sizeof( uint64_t ) );
    
    /* Initialize OTF2 related context */
    Hash_Table_init(&locations, sizeof(struct MALP_OTF2_Location_cell), 4241, MALP_OTF2_CELL_COLL_AVOID);
    Hash_Table_init(&ranks, sizeof(struct MALP_OTF2_LocationGroup_cell), 4241, MALP_OTF2_CELL_COLL_AVOID);
    Hash_Table_init(&nodes, sizeof(struct MALP_OTF2_LocationGroup_cell), 4241, MALP_OTF2_CELL_COLL_AVOID);
	
	return 0;
}

void MALP_OTF2_Event_print(void* elem)
{

    
    
}

int compare_comm(void* elem, void* to_check)
{
    return *((uint64_t*)elem) == *((uint64_t*)to_check);
}

void add_comm(uint64_t comm, uint64_t location, int only_push)
{
    uint64_t* elem = (uint64_t*)Chained_List_get_elem(&communicators, compare_comm, &comm);
    
    if(!elem)
    {
        Chained_List_push(&communicators, &comm);
        if(only_push == 0)
        {
            struct MALP_OTF2_Location* ldesc = MALP_OTF2_Location_get_by_id(location);
            OTF2_IdMap* mpi_comm_map = OTF2_IdMap_Create( OTF2_ID_MAP_SPARSE, 1 );
            //Map everything to MPI_COMM_WORLD ...
            OTF2_IdMap_AddIdPair( mpi_comm_map, comm, 0 );
            OTF2_DefWriter_WriteMappingTable(ldesc->def_writer,OTF2_MAPPING_COMM,mpi_comm_map );
        }
    }
}



//************************************************************//
//          Event Writing and collection                      //
//************************************************************//


/** TO PORT BACK
void real_register_mpi_event(struct MALP_OTF2_Event* event)
{
    struct MALP_OTF2_Location* ldesc = MALP_OTF2_Location_get_by_id(event->event.tid);
    struct MALP_OTF2_Location* source_or_dest = NULL;
    struct MPI_Event mpi_event = event->event.event.mpi_event;
    
    uint64_t comm = mpi_event.comm;
    //    uint64_t comm = 0;
    switch(mpi_event.type)
    {
    case T_MPI_SEND:
        source_or_dest = MALP_OTF2_Location_get_by_rank(mpi_event.dest);
        OTF2_EvtWriter_MpiSend(ldesc->evt_writer, NULL, event->timestamp, source_or_dest->id, comm, mpi_event.tag, mpi_event.size);
        add_comm(comm, ldesc->id, 0);
        ldesc->event_count++;
        break;
    case T_MPI_RECV:
        source_or_dest = MALP_OTF2_Location_get_by_rank(mpi_event.source);
        OTF2_EvtWriter_MpiRecv(ldesc->evt_writer, NULL, event->timestamp, source_or_dest->id, comm, mpi_event.tag, mpi_event.size);
        add_comm(comm, ldesc->id, 0);
        ldesc->event_count++;
    default:
        break;
    }
}
*/


static inline int OTF2_Analysis_MPI_Implemented( struct MALP_Trace_Event * event)
{	
	MALP_Trace_event_type etype = event->type;
	
	if( etype != MALP_EVENT_MPI )
		return 0;
	
	MPI_Event_type mpi_type = event->event.mpi_event.type;
	
    switch(mpi_type)
    {
	//    case T_MPI_SEND:
	//    case T_MPI_RECV:
	//        return 1;
	//        break;
		default:
			return 0;
    }
    return 0;
}



void OTF2_Analysis_push(struct MALP_Trace_Event *event, struct Collector *coll )
{
    struct MALP_OTF2_Location* ldesc = MALP_OTF2_Location_get_by_id(event->tid);
    
    if( !ldesc )
    {
		printf("Could not locate associated location %ld\n", event->tid );
		return;
	}
    
    //OTF_Event_print(event);
    
    struct OTF_Event *otf_event = &event->event.otf;
	
   //MALP_OTF2_Event_print( event );
    
    ldesc->last_timestamp = event->timestamp;
    
    if(otf_event->type == OTF_EVENT_ENTER)
    {
		if( OTF2_Analysis_MPI_Implemented( event ) )
        {
           //real_register_mpi_event(otf_event, event);
        }
        else
        {
			OTF2_EvtWriter_Enter( ldesc->evt_writer, NULL, event->timestamp, otf_event->str_id );
		}
        
    } else if(otf_event->type == OTF_EVENT_LEAVE )
    {
        OTF2_EvtWriter_Leave( ldesc->evt_writer, NULL, event->timestamp, otf_event->str_id );
    } else
    {
        return;
    }
    
    ldesc->event_count++;
    postFlushTimeStamp = event->timestamp; 
    
}


void OTF2_Analysis_push_meta(struct MALP_Trace_Event *evt, struct Collector *coll )
{
        struct MALP_OTF2_Location ldesc;
        struct MALP_OTF2_Location* pushed_ldesc;
        
        struct MALP_Trace_Event otf_event;
        /* Prepare event content */
       	memcpy( &otf_event, evt, sizeof( struct MALP_Trace_Event ) );
        
        if(evt->event.stream_descriptor.desc_type == DESC_BEGIN)
        {
			/* Initialize and push the new location */
			MALP_OTF2_Location_init( &ldesc, evt );
			
			/* Store this location */
			Hash_Table_push_unique(&locations, &ldesc, ldesc.id, NULL, MALP_OTF2_Location_compare);

			/* Format an enter event */
			OTF_Event_forge( &otf_event, OTF_EVENT_ENTER, MALP_EVENT_OTF, MALP_OTF_ANALYSER_FAKE_FUNCTION );
		}
		else if( evt->event.stream_descriptor.desc_type == DESC_END )
		{
			/* Format a leave event */
			OTF_Event_forge( &otf_event, OTF_EVENT_LEAVE, MALP_EVENT_OTF, MALP_OTF_ANALYSER_FAKE_FUNCTION );
		}
		
		//OTF2_push_event( &otf_event , coll );
}

void OTF2_push_event(struct MALP_Trace_Event *event, void *pcoll)
{
	struct Collector *coll = (struct Collector *)pcoll;

    switch( event->type )
    {
    case MALP_EVENT_OTF:
        OTF2_Analysis_push(event, coll );
        break;
    case MALP_EVENT_META:
        OTF2_Analysis_push_meta(event, coll );
        break;
    default:
        break;
    }
}

//************************************************************//
//          Meta-Data reduction  process                      //
//************************************************************//

int  OTF2_reduce( struct Collector * coll )
{
    int my_rank, i, j;    
    
    PMPI_Comm_rank(VMPI_Get_partition_comm(), &my_rank);
    
    PMPI_Barrier(VMPI_Get_partition_comm());
    //Send locations to rank 0
    
    if(my_rank != 0)
    {
        int number = Hash_Table_count_by_walk(&locations);
        PMPI_Send(&number, 1, MPI_INT, 0, 4042, VMPI_Get_partition_comm());
        Hash_Table_walkthrought(&locations, MALP_OTF2_Location_send_to_root);
        
    } else {
        //gather others...
        for(i=1;i<VMPI_Get_desc()->size;i++)
        {   
            int number;
            PMPI_Recv(&number, 1, MPI_INT, i, 4042, VMPI_Get_partition_comm(), NULL);
            for(j=0;j<number;j++)
            {
                MALP_OTF2_Location_recv_from_rank(i);
            }
        }
    }
    
    //Send Communicators to rank 0
    if(my_rank != 0)
    {
        uint64_t i;
        //Send the number of comms
        int number = Chained_List_count(&communicators);
        struct chained_list_view clv;
        chained_list_view_init(&clv);
        
        PMPI_Send(&number, 1, MPI_INT, 0, 4142, VMPI_Get_partition_comm());
        
        for(i=0;i<number;i++)
        {
            uint64_t* tosend = Chained_List_get_nth_seq(&communicators, &clv, i);
            PMPI_Send(tosend, 1, MPI_UINT64_T, 0, 4142, VMPI_Get_partition_comm());
        }
    } else {
        for(i=1;i<VMPI_Get_desc()->size;i++)
        {   
            int number;
            
            PMPI_Recv(&number, 1, MPI_INT, i, 4142, VMPI_Get_partition_comm(), NULL);
            
            for(j=0;j<number;j++)
            {
                uint64_t comm;
                PMPI_Recv(&comm, 1, MPI_UINT64_T, i, 4142, VMPI_Get_partition_comm(), NULL);
                add_comm(comm, 0, 1);
            }
        }
    }
    //get the highest value
    PMPI_Allreduce(&postFlushTimeStamp, &postFlushTimeStamp, 1, MPI_UINT64_T, MPI_MAX, VMPI_Get_partition_comm());
    
	return 0;
}


void register_functions(OTF2_GlobalDefWriter* global_def_writer, uint64_t base, uint64_t number, const char* const* strings, enum OTF2_Paradigm_enum paradigm )
{
    int i=0;
    for(i=0;i<number;i++)
    {
        int index = i + base;
        OTF2_GlobalDefWriter_WriteString( global_def_writer,
                                          index , /* region id*/
                                          strings[i] /* region/function name */
                                          );
        
        OTF2_GlobalDefWriter_WriteRegion( global_def_writer,
                                          index, /* region id*/
                                          index, /* function name*/
                                          index, /* alternative function name*/
                                          index, /*no description*/
                                          OTF2_REGION_ROLE_FUNCTION, 
                                          paradigm, 
                                          OTF2_REGION_FLAG_NONE, 
                                          0, 
                                          0, 
                                          0 );
    }
}

struct TMP_register_Comm {
    uint64_t* index;
    OTF2_GlobalDefWriter* global_def_writer;
    uint64_t comm_written;
};

void register_comm(void* elem, void* arg)
{
    struct TMP_register_Comm* s = (struct TMP_register_Comm*)arg;
    uint64_t comm = *(uint64_t*)elem;
    uint64_t index = *(s->index);
    OTF2_GlobalDefWriter* global_def_writer = s->global_def_writer;
    
    if(s->comm_written == 0)
    {
        char buffer[80];  
        sprintf(buffer,"MPI_COMM_WORLD");
        OTF2_GlobalDefWriter_WriteString( global_def_writer, index, buffer );
        OTF2_GlobalDefWriter_WriteComm(global_def_writer, 0, index, 0, OTF2_UNDEFINED_COMM);
        (*s->index)++;
        s->comm_written = 1;
    }
}

void register_location(void* elem, void* arg)
{
    struct MALP_OTF2_Location* ldesc = (struct MALP_OTF2_Location*)elem;
    struct MALP_OTF2_LocationGroup node;
    struct MALP_OTF2_LocationGroup rank;
    
    OTF2_GlobalDefWriter* global_def_writer = OTF2_Archive_GetGlobalDefWriter( archive );
    
    uint64_t* last_index = (uint64_t*)arg;
    char buffer[80];  
    int str_index = *last_index;
    
    node.id = ldesc->node_id;
    strcpy(node.name, ldesc->stream.hostname);
    
    //First register node if needed
    if(Hash_Table_get(&nodes, node.id, NULL, MALP_OTF2_LocationGroup_compare)==NULL)
    {
        Hash_Table_push(&nodes, &node, node.id);
        OTF2_GlobalDefWriter_WriteString( global_def_writer, str_index, node.name);
        OTF2_GlobalDefWriter_WriteSystemTreeNode(global_def_writer, node.id, str_index++, MALP_OTF2_ANALYSER_NODE_CLASS_STRING, 0 );
    }
    
    rank.id = ldesc->rank_id;
    sprintf(rank.name, "Rank %d", ldesc->stream.mpi_rank);
    
    //Then register rank if needed
    if(Hash_Table_get(&ranks, rank.id, NULL, MALP_OTF2_LocationGroup_compare)==NULL)
    {
        Hash_Table_push(&ranks, &rank, rank.id);
        OTF2_GlobalDefWriter_WriteString( global_def_writer, str_index, rank.name);
        OTF2_GlobalDefWriter_WriteLocationGroup(global_def_writer, rank.id, str_index++, OTF2_LOCATION_GROUP_TYPE_PROCESS, node.id);
    }
    
    //then register this location
    sprintf(buffer,"thread %d",ldesc->stream.omp_thread);
    OTF2_GlobalDefWriter_WriteString( global_def_writer, str_index, buffer );
    OTF2_GlobalDefWriter_WriteLocation( global_def_writer, ldesc->id, str_index, OTF2_LOCATION_TYPE_CPU_THREAD, ldesc->event_count, rank.id );
    *last_index = ++str_index;
}

static MALP_OTF2_Location_get_ids_index = 0;
void MALP_OTF2_Location_get_ids(void* elem, void* arg)
{
    struct MALP_OTF2_Location* ldesc = (struct MALP_OTF2_Location*)elem;
    uint64_t* locs = (uint64_t*)arg;
    locs[MALP_OTF2_Location_get_ids_index++] = ldesc->id;
}





int  OTF2_accumulate( struct Collector * coll )
{
	uint64_t offset = MALP_OTF2_ANALYSER_INFO_COUNT + T_MPI_COUNT + T_POSIX_COUNT;
    uint64_t i=0, j=0;
    
    int rank, size;
    PMPI_Comm_rank( VMPI_Get_partition_comm(), &rank );

	if( rank )
		return 0;

    
    /* Called before releasing */
    VMPI_Partition_desc *pdesc = VMPI_Get_desc_by_id(coll->partition_id);
    
    if( !pdesc )
    {
        printf("Could not retrieve partition desc\n");
        abort();
    }
    
    OTF2_GlobalDefWriter* global_def_writer = OTF2_Archive_GetGlobalDefWriter( archive );
    //Set the time scale
    OTF2_GlobalDefWriter_WriteClockProperties( global_def_writer, Process_timer_second, 0, Timer_tsc() );
    
    //Register global strings
    OTF2_GlobalDefWriter_WriteString( global_def_writer, MALP_OTF2_ANALYSER_EMPTY_STRING, "" );
    OTF2_GlobalDefWriter_WriteString( global_def_writer, MALP_OTF2_ANALYSER_NODE_CLASS_STRING, "Node" );
    OTF2_GlobalDefWriter_WriteString( global_def_writer, MALP_OTF2_ANALYSER_ROOT_NODE, pdesc->program_name );
    OTF2_GlobalDefWriter_WriteString( global_def_writer, MALP_OTF2_ANALYSER_PROGRAM_CLASS_STRING, "MALP Instrumented program" );
    OTF2_GlobalDefWriter_WriteString( global_def_writer, MALP_OTF2_ANALYSER_FAKE_FUNCTION, "Malp_Analysis" );
    
    OTF2_GlobalDefWriter_WriteRegion(global_def_writer, 
                                     MALP_OTF2_ANALYSER_FAKE_FUNCTION, 
                                     MALP_OTF2_ANALYSER_FAKE_FUNCTION,
                                     MALP_OTF2_ANALYSER_FAKE_FUNCTION,
                                     MALP_OTF2_ANALYSER_FAKE_FUNCTION,
                                     OTF2_REGION_ROLE_FUNCTION, 
                                     OTF2_PARADIGM_MEASUREMENT_SYSTEM, 
                                     OTF2_REGION_FLAG_NONE, 
                                     0, 
                                     0, 
                                     0);
    
    OTF2_GlobalDefWriter_WriteSystemTreeNode(global_def_writer, 0, MALP_OTF2_ANALYSER_ROOT_NODE, MALP_OTF2_ANALYSER_PROGRAM_CLASS_STRING, OTF2_UNDEFINED_SYSTEM_TREE_NODE );
    //Register instrumented functions
    
    //MPI
    register_functions(global_def_writer, MALP_OTF2_ANALYSER_INFO_COUNT, T_MPI_COUNT, MPI_Event_names, OTF2_PARADIGM_MPI);
    
    //Posix
    register_functions(global_def_writer, MALP_OTF2_ANALYSER_INFO_COUNT + T_MPI_COUNT, T_POSIX_COUNT, Wrapped_symbol_names, OTF2_PARADIGM_UNKNOWN);
    
    uint64_t last_index = offset;
    
    ++last_index;

    Hash_Table_walkthrought_2p(&locations, &last_index, register_location);
    

    struct chained_list_view clv;
    chained_list_view_init(&clv);
    
    uint64_t nl = Hash_Table_count_by_walk(&locations);
    uint64_t* locs = (uint64_t*)malloc(nl*sizeof(uint64_t));
    if( !locs )
	{
		perror("malloc");
		abort();
	}
    MALP_OTF2_Location_get_ids_index = 0;
    Hash_Table_walkthrought_2p(&locations, locs, MALP_OTF2_Location_get_ids);
    OTF2_GlobalDefWriter_WriteGroup( global_def_writer,
                                     1,
                                     MALP_OTF2_ANALYSER_EMPTY_STRING,
                                     OTF2_GROUP_TYPE_COMM_LOCATIONS,
                                     OTF2_PARADIGM_MPI,
                                     OTF2_GROUP_FLAG_NONE,
                                     nl,
                                     locs );
    free(locs);
    
    chained_list_view_init(&clv);
    uint64_t nc = Chained_List_count(&communicators);
    uint64_t* comms = (uint64_t*)malloc(nc*sizeof(uint64_t));
    
    if( !comms )
	{
		perror("malloc");
		abort();
	}
	
    for(i=0;i<nc;i++)
    {
        comms[i] = *(uint64_t*)Chained_List_get_nth_seq(&communicators, &clv, i);
    }
    
    OTF2_GlobalDefWriter_WriteGroup( global_def_writer,
                                     0,
                                     MALP_OTF2_ANALYSER_EMPTY_STRING,
                                     OTF2_GROUP_TYPE_COMM_GROUP,
                                     OTF2_PARADIGM_MPI,
                                     OTF2_GROUP_FLAG_NONE,
                                     nc,
                                     comms );
                                     
    free(comms);
    struct TMP_register_Comm tmp;
    tmp.global_def_writer = global_def_writer;
    tmp.index = &last_index;
    tmp.comm_written = 0;
    Chained_List_walkthrough_arg(&communicators, register_comm, &tmp);
	return 0;
}

//************************************************************//
//          Collector - Initialization and release            //
//************************************************************//

void OTF2_init(struct Collector *coll, uint64_t partition_id)
{
    /* Disable module according to config */
    if ( TA_Conf_get_key_bool( TA_CONF_OTF2_ANALYSIS_ENABLED ) == 0 ) {
        return;
    }
    
	/* Initialize collector */
	Collector_init( coll, MALP_EVENT_ANY,
					partition_id,
					"OTF2 Collector",
					OTF2_init_coll,
					OTF2_release_coll,
					OTF2_push_event,
					OTF2_reduce,
					OTF2_accumulate );
	
}


int  OTF2_release_coll( struct Collector * coll )
{
	Chained_List_release(&communicators);
    Hash_Table_release(&locations, MALP_OTF2_Location_release);
    Hash_Table_release(&ranks, NULL);
    Hash_Table_release(&nodes, NULL);
    PMPI_Barrier(VMPI_Get_partition_comm());
    OTF2_Archive_Close( archive );
    OTF_Event_forge_release();
	return 0;
}

