#ifndef MALP_OTF2_H
#define MALP_OTF2_H

#include <stdint.h>

#include "MALP_Config.h"
#include "Event_Desc.h"
#include "Collector.h"
#include "Wrapped_Symbols.h"
#include "Hash_Table.h"
#include "Chained_List.h"
#include "Buffered_LIFO.h"
#include "Stream_Descriptor.h"
#include "VMPI.h"

#include <otf2/OTF2_Archive.h>


//************************************************************//
//                    DATA structures                         //
//************************************************************//

/**
 * @ingroup OTF2_Internal_
 * @brief OTF2 id enumeration for basic information on archive
 */
typedef enum {
    MALP_OTF2_ANALYSER_EMPTY_STRING = 0,
    MALP_OTF2_ANALYSER_NODE_CLASS_STRING,
    MALP_OTF2_ANALYSER_ROOT_NODE,
    MALP_OTF2_ANALYSER_PROGRAM_CLASS_STRING,
    MALP_OTF2_ANALYSER_FAKE_FUNCTION,

    MALP_OTF2_ANALYSER_INFO_COUNT
} MALP_OTF2_Analyser_Info_id;


/**
 * @ingroup OTF2_Internal_
 * @brief Used to keep informations about locations and share it between analyzer MPI processes. 
 * When shared via MPI, evt_writer and def_writer are updated according to the 
 * id and the FIFOs are not, but they are not necessary at this point. 
 */
struct MALP_OTF2_Location
{
    uint64_t                    id;                 /**< @brief the location id (same as stream.tid) */
    uint64_t                    rank_id;            /**< @brief the rank */
    uint64_t                    node_id;            /**< @brief the node id (identifies the physical node) */
    uint64_t                    last_timestamp;     /**< @brief timestamp of last event */
    uint64_t                    event_count;        /**< @brief count of events in this location */
    
    struct Stream_Descriptor    stream;             /**< @brief the stream descriptor */
    
    OTF2_EvtWriter*             evt_writer;         /**< @brief the event writer */
    OTF2_DefWriter*             def_writer;         /**< @brief the def writer */
};

/**
 * @ingroup OTF2_Internal_
 * @brief Maximum size of a LocationGroup name
 */
#define MAX_NAME_SIZE 32

/**
 * @ingroup OTF2_Internal_
 * @brief Defines a location group
 */
struct MALP_OTF2_LocationGroup
{
    uint64_t id;                /**< @brief the id of the group */
    char name[MAX_NAME_SIZE];   /**< @brief the name of the group */
};

/**
 * @ingroup OTF2_Internal_
 * @brief Coll avoid used for the hashtables
 */
#define MALP_OTF2_CELL_COLL_AVOID 10

/**
 * @ingroup OTF2_Internal_
 * @brief Cell to initialize the MALP_OTF2_Location Hash Table
 */
struct MALP_OTF2_Location_cell
{
	struct Hash_Table_Internal ___;
	struct MALP_OTF2_Location ent[MALP_OTF2_CELL_COLL_AVOID];
};

/**
 * @ingroup OTF2_Internal_
 * @brief Cell to initialize the MALP_OTF2_LocationGroup Hash Table
 */
struct MALP_OTF2_LocationGroup_cell
{
	struct Hash_Table_Internal ___;
	struct MALP_OTF2_Location ent[MALP_OTF2_CELL_COLL_AVOID];
};

//************************************************************//
//          Collector - Analysis component functions          //
//************************************************************//

/**
 * @ingroup OTF2_Analysis_
 * @brief Basically tests if the collector must be launched and sets the corresponding callbacks
 * @param coll the measure collector
 * @param partition_id the id of the instrumented partition
 */
void OTF2_init(struct Collector *coll, uint64_t partition_id);
//void OTF2_Analysis_init( struct Measure_Collector *mc, uint64_t partition_id );

/**
 * @ingroup OTF2_Analysis_
 * @brief Sets the OTF2 environment up by initializing static data
 * @param coll the measure collector
 */
int  OTF2_init_coll( struct Collector * coll );
//void OTF2_Analysis_setup( struct Measure_Collector *mc );

/**
 * @ingroup OTF2_Analysis_
 * @brief Call a push function for an event depending on the type of event to store (Meta, Posix, MPI)
 * @param evt the event to be pushed
 * @param coll the measure collector
 */
void OTF2_push_event(struct MALP_Trace_Event *, void *);
//void OTF2_Analysis_push( struct MALP_Trace_Event *evt, struct Measure_Collector *mc );

/**
 * @ingroup OTF2_Analysis_
 * @brief Gathers data from other ranks to prepare rendering
 * @param coll the measure collector
 */
int  OTF2_reduce( struct Collector * coll );
//void OTF2_Analysis_reduce( struct Measure_Collector *mc );

/**
 * @ingroup OTF2_Analysis_
 * @brief Registers everything into OTF2 
 * @param coll the measure collector
 */
int  OTF2_accumulate( struct Collector * coll );
//void OTF2_Analysis_render( struct Measure_Collector *mc );

/**
 * @ingroup OTF2_Analysis_
 * @brief Releases the structures
 * @param coll the measure collector
 */
int  OTF2_release_coll( struct Collector * coll );
//void OTF2_Analysis_unset( struct Measure_Collector *mp );


//***************************OTF2*****************************//

/**
 * @ingroup OTF2_Specific_
 * @brief Function called before the OTF2_Archive flush (before OTF2_Archive closing) and stored in the OTF2_FlushCallbacks structure of an OTF2_Archive archive
 * @param userData
 * @param fileType otf/evt/def file type
 * @param location
 * @param callerData
 * @param final
 * @return
 */
OTF2_FlushType pre_flush(   void* userData,
                            OTF2_FileType fileType,
                            OTF2_LocationRef location,
                            void* callerData,
                            bool final );
/**
 * @ingroup OTF2_Specific_
 * @brief Function called after the @link OTF2_Archive @endlink flush (after OTF2_archive closing) and stored in the OTF2_FlushCallbacks structure of an OTF2_Archive archive
 * @param userData
 * @param fileType otf/evt/def file type
 * @param location
 * @return
 */
OTF2_TimeStamp post_flush(  void* userData,
                            OTF2_FileType fileType,
                            OTF2_LocationRef location );


//*************************Internal***************************//

/**
 * @ingroup OTF2_Internal_
 * @brief comparison function for MALP_OTF2_LocationGroup. 
 * @param elem the MALP_OTF2_LocationGroup
 * @param id the id to compare
 * @param extra [not used]
 * @return 1 if the id of elem is the same as id
 */
int MALP_OTF2_LocationGroup_compare(void* elem, uint64_t id, void* extra);

/**
 * @ingroup OTF2_Internal_
 * @brief comparison function for MALP_OTF2_Location. 
 * @param elem the MALP_OTF2_Location
 * @param id the id to compare
 * @param extra [not used]
 * @return 1 if the id of elem is the same as id
 */
int MALP_OTF2_Location_compare(void* elem, uint64_t id, void* extra);

/**
 * @ingroup OTF2_Internal_
 * @brief Initializes a MALP_OTF2_Location. 
 * @param elem the location to initialize
 * @param evt Optionnal pointer to a metadata event from which to inherit ctx
 */
void MALP_OTF2_Location_init(struct MALP_OTF2_Location* loc, struct MALP_Trace_Event *evt);

/**
 * @ingroup OTF2_Internal_
 * @brief Printer function for a MALP_OTF2_Location
 * @param elem the element to print
 */
void MALP_OTF2_Location_print(void* elem);

/**
 * @ingroup OTF2_Internal_
 * @brief Sends a MALP_OTF2_Location to a given rank via MPI
 * @param elem the MALP_OTF2_Location to send
 * @param rank the rank to send to
 */
void MALP_OTF2_Location_send_to_rank(void* elem, int rank);

/**
 * @ingroup OTF2_Internal_
 * @brief Retrieve a MALP_OTF2_Location from a given rank via MPI and register it to the local locations
 * @param rank the wanted rank
 */
void MALP_OTF2_Location_recv_from_rank(int rank);

/**
 * @ingroup OTF2_Internal_
 * @brief Sends a MALP_OTF2_Location to rank 0 via MPI
 * @param elem the MALP_OTF2_Location to send
 */
void MALP_OTF2_Location_send_to_root(void* elem);

/**
 * @ingroup OTF2_Internal_
 * @brief Retrieves a MALP_OTF2_Location corresponding to a given id. 
 * @warning If there is no MALP_OTF2_Location with this id, it will exit(1)
 * @param tid the id of the wanted MALP_OTF2_Location
 * @return the corresponding MALP_OTF2_Location
 */
struct MALP_OTF2_Location* MALP_OTF2_Location_get_by_id(uint64_t tid);

/**
 * @ingroup OTF2_Internal_
 * @brief Used in a Hash_Table_walkthrought_3p to find a location based on its rank
 * @param elem  The MALP_OTF2_Location from the Hash_Table
 * @param second the wanted rank (uint64_t*)
 * @param third the destination MALP_OTF2_Location**
 */
void MALP_OTF2_Location_find_by_rank(void* elem, void* second, void* third);

/**
 * @ingroup OTF2_Internal_
 * @brief Retrieves a MALP_OTF2_Location corresponding to a given rank. 
 * @warning If there is no MALP_OTF2_Location with this rank, it will exit(1)
 * @warning If there is more than one MALP_OTF2_Location with this rank, it will get the latest found
 * @param rank the rank of the wanted MALP_OTF2_Location
 * @return the corresponding MALP_OTF2_Location
 */
struct MALP_OTF2_Location* MALP_OTF2_Location_get_by_rank(uint64_t rank);

/**
 * @ingroup OTF2_Internal_
 * @brief used in Chained_List_get_elem to compare stored communicators
 * @param elem the Chained_List element
 * @param compare the wanted value
 * @return 1 if they are the same
 */
int compare_comm(void* elem, void* compare);

/**
 * @ingroup OTF2_Internal_
 * @brief Adds a communicator to local storage
 * @param comm the communicator to add
 * @param location the location id the communicator belongs to
 * @param only_push if 1 => only store in local storage, if 0 => map it to MPI_COMM_WORLD in OTF2
 */
void add_comm(uint64_t comm, uint64_t location, int only_push);


/**
 * @ingroup OTF2_Internal_
 * @brief Must indicate if the event has specific treatment in OTF2 (such as MPI_SEND etc.)
 * If the event is specific, events are registered like this : 
 *   - Enter
 *   - Specific
 *   - Leave
 * @warning No event is specific for now...
 * @param type the type of event
 * @return 1 if it is specific
 */
int is_mpi_specific(MPI_Event_type type);


/**
 * @ingroup OTF2_Internal_
 * @brief Register all pending leave events (ordered) which are older than a timestamp
 * @param enter_timestamp the timestamp to compare to
 * @param tid the MALP_OTF2_Location id
 */
void register_pending_leave_events(uint64_t enter_timestamp, uint64_t tid);

/**
 * @ingroup OTF2_Internal_
 * @brief Registers a set of functions (Regions) into OTF2
 * @param global_def_writer the OTF2 global writer
 * @param base Where to start with string index
 * @param number the number of functions to register
 * @param strings the strings (functions names)
 * @param paradigm the paradigm of the set of functions 
 */
void register_functions(OTF2_GlobalDefWriter* global_def_writer, uint64_t base, uint64_t number, const char* const* strings, enum OTF2_Paradigm_enum paradigm );

/**
 * @ingroup OTF2_Internal_
 * @brief Registers a communicator into OTF2
 * @warning For now, everything is registered as MPI_COMM_WORLD. 
 * @param elem the communicator
 * @param arg a temporary structure to control the process
 */
void register_comm(void* elem, void* arg);

/**
 * @ingroup OTF2_Internal_
 * @brief Registers a location into OTF2
 * It also registers (if needed) : 
 *   - The rank location group
 *   - The physycal node (system tree node)
 * @param elem the location to register
 * @param arg the string index
 */
void register_location(void* elem, void* arg);

/**
 * @ingroup OTF2_Internal_
 * @brief Retrieves all the MALP_OTF2_Location ids available and store them into arg. used with Hash_Table_walkthrought_2p
 * @param elem the Hash_Table element
 * @param arg where to store ids (uint64_t*)
 */
void MALP_OTF2_Location_get_ids(void* elem, void* arg);

/**
 * @ingroup OTF2_Internal_
 * @brief Splits a MALP event into Enter/Leave events and send them to the registering system. 
 * @param event the new MALP event 
 */
void OTF2_Analysis_generic_push(struct MALP_Trace_Event *event );

/**
 * @ingroup OTF2_Analysis_
 * @brief Creates a location descriptor corresponding to source Stream and register it into locations
 * The location descriptor is only created if the MALP_EVENT_META event is a DESC_BEGIN one.
 * @param evt The MALP_EVENT_META event
 * @param mc The measure collector
 */
void OTF2_Analysis_push_meta(struct MALP_Trace_Event *evt, struct Collector *coll );

//************************************************************//

#endif
