#include "Rank_Matrix.h"

#include <math.h>

#include "MPI_Events.h"
#include "Meta_Data_Handler.h"
#include "Shared_Cache.h"
#include "Timer.h"



struct Array * Rank_Matrix_init_profiling_entry(int size)
{
		struct Array * entry = calloc( sizeof( struct Array ) , 3 );
		
		if( !entry )
		{
			perror("calloc");
			return NULL;
		}
		
		/* TODO fetch from config */
		int is_sparse = 0;
		
		if( Array_init_1D( &entry[0], size, is_sparse ) )
		{
			free( entry );
			return NULL;
		}
		
		if( Array_init_1D( &entry[1], size, is_sparse ) )
		{
			free( entry );
			return NULL;
		}
		
		if( Array_init_1D( &entry[2], size, is_sparse ) )
		{
			free( entry );
			return NULL;
		}
		
		return entry;
}

void Rank_Matrix_release_profiling_entry( struct Array * entry )
{
	if(!entry)
		return;
	
	Array_release( &entry[0] );
	Array_release( &entry[1] );
	Array_release( &entry[2] );
	
	free( entry );
}

void Rank_Matrix_reduce_profiling_entry( struct Array * entry, MPI_Op op )
{
	if(!entry)
		return;

	Array_reduce( &entry[0], op );
	Array_reduce( &entry[1], op );
	Array_reduce( &entry[2], op );
}


json_t * Rank_Matrix_serialize_array( struct Array *array )
{
	double square_side = ceil( sqrt( array->Dx) );
	uint64_t offset = 0;
	
	json_t * ret = json_object();
	
	/* Fill ctx */
	json_t * min = json_real( Array_get_min(array) );
	json_t * max = json_real( Array_get_max(array) );
	
	json_object_set( ret, "min", min );
	json_object_set( ret, "max", max );
	
	json_decref( min );
	json_decref( max );
	
	/* Fill data */
	json_t * entry_data = json_array();
	
	int i,j;
	
	for( i = 0 ; i < square_side ; i++ )
	{
		/* Create line */
		json_t * line = json_array();
		
		for( j = 0 ; j < square_side ; j++ )
		{

			json_t * jval = NULL;
			
			if( offset < array->Dx )
			{
				double val = Array_get_1D (array, offset);
				jval = json_real( val );
			}
			else
			{
				/* Here we pad with nulls if we do not have
				 * a perfect square ... */
				jval = json_null();
			}
			
			json_array_push( line, jval );
			json_decref( jval );
			
			offset++;		
		}
		
		/* Push line */
		json_array_push( entry_data, line );
		json_decref( line );
	}
	
	json_object_set( ret, "map", entry_data );
	json_decref( entry_data );

	return ret;
}

json_t * Rank_Matrix_serialize_profiling_entry( struct Array *entries, char *name )
{
	json_t * entry_obj = json_object();
	
	json_t * hits = Rank_Matrix_serialize_array(&entries[0]);
	json_t * time = Rank_Matrix_serialize_array(&entries[1]);
	
	
	json_t * size = NULL;
	
	if( Array_has_datas(&entries[2]) )
	{
		size = Rank_Matrix_serialize_array(&entries[2]);
	}
	else
	{
		size = json_null();
	}

	json_object_set( entry_obj, "hits", hits );
	json_object_set( entry_obj, "time", time );
	json_object_set( entry_obj, "size", size );
	
	
	json_t * jname = json_string( name );
	
	json_object_set( entry_obj, "name", jname );
	
	json_decref( jname );
	
	json_decref( hits );
	json_decref( time );
	json_decref( size );
	
	return entry_obj;
}


struct Array * Rank_Matrix_get_profiling_entry( struct Rank_Matrix *rm , MALP_Trace_event_type type, int id )
{
	struct Array ** target;
	
	int expected_size = 0;
	
	switch( type )
	{
		case MALP_EVENT_MPI:
		
			if( T_MPI_COUNT <= id )
				return NULL;
		
			target = rm->mpi;
		break;
		case MALP_EVENT_WRAPPED:
		
			if( T_POSIX_COUNT <= id )
				return NULL;
		
			target = rm->posix;
		break;
		default:
			target = NULL;
		break;
	}
	
	if( ! target )
		return;
	
	if( ! target[id] )
	{
		target[id] = Rank_Matrix_init_profiling_entry(rm->size);
	}
	
	return target[id];
}



int  Rank_Matrix_init_coll( struct Collector * coll )
{
	/* Init collector internal */
	coll->internal_data = malloc( sizeof( struct Rank_Matrix ) );
	
	if(! coll->internal_data )
	{
		perror("malloc");
		abort();
	}
	
	struct Rank_Matrix *rm = (struct Rank_Matrix *)coll->internal_data;
	
	/* Retrieve partition size */
	rm->size = Collector_partition_size( coll );
	
	
	/* Allocate accumulators */
	
	/* MPI */
	rm->mpi = calloc( sizeof( struct Array * ), T_MPI_COUNT );
	
	if( !rm->mpi )
	{
		perror("calloc");
		abort();
	}
	
	/* POSIX */
	rm->posix = calloc( sizeof( struct Array * ), T_POSIX_COUNT );
	
	if( !rm->posix )
	{
		perror("calloc");
		abort();
	}	
	
	
	return 0;
}


int  Rank_Matrix_release_coll( struct Collector * coll )
{
	struct Rank_Matrix *rm = (struct Rank_Matrix *)coll->internal_data;
	
	/* Free accumulators */
	int i;
	
	/* MPI */
	for( i = 0 ; i < T_MPI_COUNT ; i++ )
	{
		Rank_Matrix_release_profiling_entry( rm->mpi[i] );
	}
	
		
	free( rm->mpi );
	rm->mpi = NULL;
	
	/* POSIX */
	for( i = 0 ; i < T_POSIX_COUNT ; i++ )
	{
		Rank_Matrix_release_profiling_entry( rm->posix[i] );
	}	
	
	free( rm->posix );
	rm->posix = NULL;
	
	rm->size = 0;
	
	
	free( rm );
	
	return 0;
}



void Rank_Matrix_push_event(struct MALP_Trace_Event *event, void *pcoll)
{
	struct Collector * coll = (struct Collector *)pcoll;
	
	int type = -1;
	struct Rank_Matrix *rm = (struct Rank_Matrix *)coll->internal_data;
	
	struct Array *target = NULL;
	
	struct MPI_Event *mpievt;
	struct Wrapped_Event *we;
	
	switch( event->type )
	{
		case MALP_EVENT_MPI:
			mpievt = &event->event.mpi_event;
			type = mpievt->type;
			
			target = Rank_Matrix_get_profiling_entry( rm , event->type, type );
			
			if( !target )
				return;
		
			if( mpievt->size )
				Array_inc_val_1D( &target[2], mpievt->size, event->rank );
			
			break;
		case MALP_EVENT_WRAPPED :
			we = &event->event.wrapped_event;

			type = we->type;
			
			target = Rank_Matrix_get_profiling_entry( rm , event->type, type );
		
			if( !target )
				return;
	
			if( we->size )
				Array_inc_val_1D( &target[2], we->size, event->rank );
			break;
		default:
				return;
	}

	Array_inc_1D( &target[0], event->rank );
	Array_inc_val_1D( &target[1], event->duration / Timer_second(), event->rank );
}

void Rank_Matrix_unify( struct Array ** array, int array_size, int entry_size )
{
	/* Prepare boolean arrays */
	int * allocated = calloc( sizeof( int ), array_size );
	int * allocated_out = calloc( sizeof( int ), array_size );
	
	if( !allocated || !allocated_out)
	{
		perror("calloc");
		abort();
	}
	
	/* Copy back our datas */
	int i;
	
	for( i = 0 ; i < array_size ; i++ )
	{
		if( array[i] != NULL )
			allocated[i] = 1;
	}
	
	/* Perform the allreduce */
	PMPI_Allreduce(allocated, allocated_out, array_size, MPI_INT, MPI_SUM, VMPI_Get_partition_comm());
	
	/* Make sure all processes initialized existing entries */
	for( i = 0 ; i < array_size ; i++ )
	{
		if( allocated_out[i] && !array[i] )
			array[i] = Rank_Matrix_init_profiling_entry(entry_size);
	}
	
	free( allocated );
	free( allocated_out );
}





int  Rank_Matrix_reduce( struct Collector * coll )
{
	struct Rank_Matrix *rm = (struct Rank_Matrix *)coll->internal_data;
	
	/* First we have to allocate entries in every processes
	 * if it has been hit in a single one*/
	 Rank_Matrix_unify( rm->mpi , T_MPI_COUNT, rm->size );
	 Rank_Matrix_unify( rm->posix , T_POSIX_COUNT, rm->size );
	 
	/* We can now reduce existing entries */
	
	int i;
	
	/* MPI */
	for( i = 0 ; i < T_MPI_COUNT ; i++ )
	{
		Rank_Matrix_reduce_profiling_entry( rm->mpi[i], MPI_SUM );
	}
	
	/* POSIX */
	for( i = 0 ; i < T_POSIX_COUNT ; i++ )
	{
		Rank_Matrix_reduce_profiling_entry( rm->posix[i], MPI_SUM );
	}
	
	return 0;
}





int  Rank_Matrix_accumulate( struct Collector * coll )
{
	struct Rank_Matrix *rm = (struct Rank_Matrix *)coll->internal_data;

	int rank;
	PMPI_Comm_rank( VMPI_Get_partition_comm(), &rank );

	if( rank )
		return 0;

	/* Only process 0 from here */
	
	json_t * rank_matrices = json_array();
	
	
	json_t * entry = NULL;
	
	int i;
	
	/* MPI */
	for( i = 0 ; i < T_MPI_COUNT ; i++ )
	{
		if( rm->mpi[i] )
		{
			entry = Rank_Matrix_serialize_profiling_entry( rm->mpi[i], (char *)MPI_Event_names[i] );
			json_array_push( rank_matrices, entry );
			json_decref( entry );
		}
	}
	
	/* POSIX */
	for( i = 0 ; i < T_POSIX_COUNT ; i++ )
	{
		if( rm->posix[i] )
		{
			entry = Rank_Matrix_serialize_profiling_entry( rm->posix[i], (char *)Wrapped_symbol_names[i] );
			json_array_push( rank_matrices, entry );
			json_decref( entry );
		}
	}	
	 
	char uri[100];
	snprintf(uri, 100, "%s", coll->analysis_cache_name );
	nodeSharedCache_set( &nsc, uri, rank_matrices );
	
	json_decref( rank_matrices );

	return 0;
}




void Rank_Matrix_init(struct Collector *coll, uint64_t partition_id)
{
		
	Collector_init( coll, MALP_EVENT_ANY,
					partition_id,
					"Rank_Matrix",
					Rank_Matrix_init_coll,
					Rank_Matrix_release_coll,
					Rank_Matrix_push_event,
					Rank_Matrix_reduce,
					Rank_Matrix_accumulate );
	
}
