/* ############################ MALP License ############################## */
/* # Fri Jan 18 14:00:00 CET 2013                                         # */
/* # Copyright or (C) or Copr. Commissariat a l'Energie Atomique          # */
/* #                                                                      # */
/* # This software is governed by the CeCILL-C license under French law   # */
/* # and abiding by the rules of distribution of free software.  You can  # */
/* # use, modify and/ or redistribute the software under the terms of     # */
/* # the CeCILL-C license as circulated by CEA, CNRS and INRIA at the     # */
/* # following URL http://www.cecill.info.                                # */
/* #                                                                      # */
/* # The fact that you are presently reading this means that you have     # */
/* # had knowledge of the CeCILL-C license and that you accept its        # */
/* # terms.                                                               # */
/* #                                                                      # */
/* # Authors:                                                             # */
/* #   - BESNARD Jean-Baptiste jean-baptiste.besnard@cea.fr               # */
/* #                                                                      # */
/* ######################################################################## */
#include "Red_HT.h"

#include <mpi.h>

#include "Debug.h"
#include "Binary_Tree.h"
#include "VMPI.h"


void Red_Ht_init( struct Red_Ht *rht, size_t payload_size, uint64_t table_size, uint64_t coll_avoid)
{
	Hash_Table_init( &rht->ht, payload_size, table_size, coll_avoid );
	rht->internal_elem_size = rht->ht.intern_payload_size;
}

void Red_Ht_clone(struct Red_Ht *rht, struct Red_Ht *source)
{
	Red_Ht_init( rht, source->ht.sizeof_payload, source->ht.table_size, source->ht.coll_avoid);
}



void Red_Ht_release(struct Red_Ht *rht, void (*free_f)(void *) )
{
	Hash_Table_release( &rht->ht, free_f );
	rht->internal_elem_size = 0;
}



struct Red_HT_buffer_desc
{
	void *buffer;
	uint64_t buffer_size;
	uint64_t current_offset;
	uint64_t elem_size;
};

void __Red_Ht_reduce_push_elem(void *elem, void *pbuffer_desc)
{
	struct Red_HT_buffer_desc *desc = (struct Red_HT_buffer_desc *)pbuffer_desc;

	if(desc->buffer_size <= desc->current_offset)
	{
		ERROR("Overflow when pushing to reduce buffer (%lld over %lld)\n",desc->current_offset, desc->buffer_size);
		exit(1);
	}

	memcpy((char *)desc->buffer + desc->current_offset * desc->elem_size,
		   elem,
		   desc->elem_size);

	desc->current_offset++;
}


void * Red_Ht_serialize(struct Red_Ht *rht, uint64_t *elem_count, uint64_t *elem_size)
{
	void *local_data = NULL;

	*elem_size = rht->internal_elem_size;
	*elem_count = Hash_Table_count_by_walk(&rht->ht);

	/* Build local accumulator */
	{
		local_data = malloc(*elem_count * rht->internal_elem_size);

		if(!local_data)
		{
			PERROR("malloc");
			exit(1);
		}

		struct Red_HT_buffer_desc desc = {local_data, *elem_count, 0, *elem_size};

		Hash_Table_walkthrought_2p( &rht->ht, (void*) &desc,  __Red_Ht_reduce_push_elem );
	}

	return local_data;
}

/**
 * @ingroup Red_HT_
 * @brief The Red_Ht_payload_desc struct
 */
struct Red_Ht_payload_desc
{
	uint64_t elem_count;
	uint64_t elem_size;
};


void Red_Ht_accumulate(struct Red_Ht *accumulator, int source, void (*red_func)(void * ,void *), int (*test_func)(void *, uint64_t , void *), uint64_t key_func(void *) )
{
	/* Receive acc size from child */
	struct Red_Ht_payload_desc desc = {0, 0};

	if(0<=source)
	{
		PMPI_Recv((void *)&desc, sizeof(struct Red_Ht_payload_desc), MPI_CHAR, source, 7896, VMPI_Get_partition_comm(), MPI_STATUS_IGNORE);
	}


	/* Nothing to do */
	if(desc.elem_count == 0)
		return;


	if(accumulator->internal_elem_size != desc.elem_size)
	{
		ERROR("Elem size mismatch %lld versus %lld\n", accumulator->internal_elem_size, desc.elem_size);
		exit(1);
	}

	/* Allocate child acc */
	void *child_acc = malloc(desc.elem_count * desc.elem_size);

	if(!child_acc)
	{
		PERROR("malloc");
		exit(1);
	}

	/* Receive child acc */
	PMPI_Recv((void *)child_acc, desc.elem_count * desc.elem_size, MPI_CHAR, source, 7896, VMPI_Get_partition_comm(), MPI_STATUS_IGNORE);

	/* Reduce Data */
	int i;

	for(i = 0 ; i < desc.elem_count ; i++)
	{
		void *elem = (char *)child_acc + i * desc.elem_size;
		uint64_t elem_key = key_func(elem);

		void *loc_elem = Red_Ht_Table_get( accumulator, elem_key , elem, test_func);

		/* We found the same entry locally */
		if(loc_elem)
		{
			red_func(loc_elem, elem);
		}
		else
		{
			/* No such elem, lets push it */
			Red_Ht_Table_push( accumulator, elem , key_func(elem) );
		}
	}

	free(child_acc);
}


void Red_Ht_reduce(struct Red_Ht *rht, void (*red_func)(void * ,void *), int (*test_func)(void *, uint64_t , void *), uint64_t key_func(void *), void (*free_f)(void *) )
{
	int parent, left_child, right_child;
	int rank, size;

	PMPI_Comm_rank(VMPI_Get_partition_comm(), &rank);
	PMPI_Comm_size(VMPI_Get_partition_comm(), &size);


	/* Build a btree */
	struct Binary_Tree tree;
	Binary_Tree_init(&tree, rank, size);

	parent = Binary_Tree_parent(&tree);
	left_child = Binary_Tree_left_child(&tree);
	right_child = Binary_Tree_right_child(&tree);

	Binary_Tree_release(&tree);

	/* Initialize accumulation HT */
	struct Red_Ht accumulator;
	Red_Ht_clone(&accumulator, rht);

	/* Serialize local hash_table */
	uint64_t local_elem_count = 0;
	uint64_t elem_size = 0;

	/* Push all elems to accumulation HT */
	{
		void * local_buffer = Red_Ht_serialize(rht, &local_elem_count, &elem_size);

		int i;

		for( i = 0 ; i < local_elem_count ; i++)
		{
			Red_Ht_Table_push( &accumulator, (char *)local_buffer + i * elem_size , key_func((char *)local_buffer + i * elem_size) );
		}

		free( local_buffer );
		local_buffer = NULL;
	}

	/* Reduce children data */

	Red_Ht_accumulate(&accumulator, left_child, red_func, test_func, key_func );
	Red_Ht_accumulate(&accumulator, right_child, red_func, test_func, key_func );

	/* Send reduced data to parent */
	{

		uint64_t acc_elem_count = 0;
		uint64_t acc_elem_size = 0;

		void * accumulation_buffer = Red_Ht_serialize(&accumulator, &acc_elem_count, &acc_elem_size);

		//printf("[%d parent %d  Lc %d  Rc : %d : ACC has %lld elems \n", rank, parent, left_child, right_child, acc_elem_count);

		struct Red_Ht_payload_desc desc = {acc_elem_count, acc_elem_size};

		/* Send accumulator size to parent */
		if(0<=parent)
		{
			PMPI_Send((void *)&desc, sizeof(struct Red_Ht_payload_desc), MPI_CHAR, parent, 7896, VMPI_Get_partition_comm());
		}

		/* Send data to parent if needed */
		if(0<=parent && desc.elem_count)
		{
			 PMPI_Send((void *)accumulation_buffer, acc_elem_count * acc_elem_size, MPI_CHAR, parent, 7896, VMPI_Get_partition_comm());
		}

		free(accumulation_buffer);

	}

	/* 0 Stores the result */
	if (rank == 0 )
	{
		Red_Ht_release(rht, free_f );
		memcpy(rht, &accumulator, sizeof(struct Red_Ht) );
	}
	else
	{
		Red_Ht_release(&accumulator, free_f );
	}
}
