/* ############################ MALP License ############################## */
/* # Fri Jan 18 14:00:00 CET 2013                                         # */
/* # Copyright or (C) or Copr. Commissariat a l'Energie Atomique          # */
/* #                                                                      # */
/* # This software is governed by the CeCILL-C license under French law   # */
/* # and abiding by the rules of distribution of free software.  You can  # */
/* # use, modify and/ or redistribute the software under the terms of     # */
/* # the CeCILL-C license as circulated by CEA, CNRS and INRIA at the     # */
/* # following URL http://www.cecill.info.                                # */
/* #                                                                      # */
/* # The fact that you are presently reading this means that you have     # */
/* # had knowledge of the CeCILL-C license and that you accept its        # */
/* # terms.                                                               # */
/* #                                                                      # */
/* # Authors:                                                             # */
/* #   - BESNARD Jean-Baptiste jean-baptiste.besnard@cea.fr               # */
/* #                                                                      # */
/* ######################################################################## */

/**
 * @addtogroup Bit_Array_
 * @{
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>

#ifndef BIT_ARRAY_H
#define BIT_ARRAY_H

#ifdef __cplusplus
extern "C"
{
#endif

/**
 * @brief Masks used to get and set bits. 
 *
 * It is binary anded/ored with the buffer to get/set the bit value : 
 *  @code
 *  bit = buffer & mask;
 *  buffer = buffer | mask;
 *  @endcode
 * Example in 4 bits : 
 * 0001
 * 0010
 * 0100
 * 1000
 */
static const int get_set_bit_mask[] = {
	1,
	2,
	4,
	8,
	16,
	32,
	64,
	128
};


/**
 * @brief Masks used to unset bits
 *
 *  It is binary anded to unset bit on the buffer : 
 *  @code
 *  buffer = buffer & mask;
 *  @endcode
 * Example in 4 bits : 
 * 1110
 * 1101
 * 1011
 * 0111
 */
static const int unset_bit_mask[] = {
	254,
	253,
	251,
	247,
	239,
	223,
	191,
	127
};

/**
 * @brief Defines a bit array
 */
struct Bit_Array {
	uint8_t *array;     /**< @brief The data */
	uint64_t size;      /**< @brief The number of uint8_t array contains */
	uint64_t real_size; /**< @brief The number of bits the array contains */
};

/**
 * @brief Getter on the array size
 * @param ba the array
 * @return the number of bits ba contains
 */
static inline uint64_t Bit_Array_size(struct Bit_Array *ba)
{
	return ba->real_size;
}

/**
 * @brief Bit_Array initializer
 * @param ba the Bit_Array to initialize
 * @param size the number of bits wanted fot ba
 */
void Bit_Array_init( struct Bit_Array *ba, uint64_t size );

/**
 * @brief Bit_Array deinitializer
 * @param ba the Bit_Array to deinitialize
 */
void Bit_Array_release( struct Bit_Array *ba );

/**
 * @brief Copies a Bit_Array into another
 * @param dest the destination Bit_Array
 * @param src the source Bit_Array
 *
 * @warning Both src and dest must be initialized at the same size before copying. 
 */
void Bit_Array_replicate( struct Bit_Array *dest, struct Bit_Array *src );

/**
 * @brief Prints a Bit_Array on standard output
 * @param ba the Bit_Array to print
 */
void Bit_Array_dump( struct Bit_Array *ba );

/**
 * @brief Sets a bit in a Bit_Array
 * @param ba the Bit_Array where to set the bit
 * @param key the index of the bit to set
 * @param value the value to set
 */
static inline void Bit_Array_set(  struct Bit_Array *ba, uint64_t key, uint8_t value )
{

    if( ba->real_size <= key )
    {
        //printf(" SET Out of bound (%ld)\n", key);
        return;
    }

    uint64_t extern_offset = key >> 3;
    uint8_t local_offset = key - extern_offset*8;

    uint8_t *target = &ba->array[extern_offset];

    if( value )
        *target |= get_set_bit_mask[local_offset];
    else
        *target &= unset_bit_mask[local_offset];

}

/**
 * @brief Gets a bit in a Bit_Array
 * @param ba the Bit_Array where to get the bit
 * @param key the index of the bit to get
 * @return the value of the bit (3 if out of bound)
 */
static inline uint8_t Bit_Array_get(  struct Bit_Array *ba, uint64_t key )
{
    if( ba->real_size <= key )
    {
        //printf(" GET Out of bound (%ld) \n", key);
        return 3;
    }

    uint64_t extern_offset = key >> 3;
    uint8_t local_offset = key - extern_offset*8;


    return (ba->array[extern_offset] & get_set_bit_mask[local_offset]) >> local_offset;

}

#ifdef __cplusplus
}
#endif

#endif

/**
 * @}
 */
