/* ############################ MALP License ############################## */
/* # Fri Jan 18 14:00:00 CET 2013                                         # */
/* # Copyright or (C) or Copr. Commissariat a l'Energie Atomique          # */
/* #                                                                      # */
/* # This software is governed by the CeCILL-C license under French law   # */
/* # and abiding by the rules of distribution of free software.  You can  # */
/* # use, modify and/ or redistribute the software under the terms of     # */
/* # the CeCILL-C license as circulated by CEA, CNRS and INRIA at the     # */
/* # following URL http://www.cecill.info.                                # */
/* #                                                                      # */
/* # The fact that you are presently reading this means that you have     # */
/* # had knowledge of the CeCILL-C license and that you accept its        # */
/* # terms.                                                               # */
/* #                                                                      # */
/* # Authors:                                                             # */
/* #   - BESNARD Jean-Baptiste jean-baptiste.besnard@cea.fr               # */
/* #                                                                      # */
/* ######################################################################## */

/**
 * @addtogroup Bloom_Filter_
 * @{
 */


#ifndef BLOOM_FILTER_H
#define BLOOM_FILTER_H

#include "Bit_Array.h"
#include "Spinlock.h"
#include "CRC64.h"

#ifdef __cplusplus
  extern "C"
  {
#endif

/**
 * @brief This is the main structure defining the Bloom Filter. 
 */
struct Bloom_Filter
{
    struct Bit_Array array; /**< @brief The bit Array containing the filter data */
    MALP_Spinlock lock;     /**< @brief The lock for concurrent access on the bit array */
    uint64_t size;          /**< @brief The size of the filter data */
    size_t sizeof_payload;  /**< @brief The size of elements that can be added to the filter */
};

/**
 * @brief Initializes a Bloom_Filter. 
 * @param bf the filter to initialize
 * @param size the size of the filter's internal bit array
 * @param sizeof_payload the size of data that can be added to the filter
 */
void Bloom_Filter_init( struct Bloom_Filter *bf , uint64_t size, size_t sizeof_payload );

/**
 * @brief Releases a Bloom_Filter. 
 * @param bf the filter to be freed
 */
void Bloom_Filter_release( struct Bloom_Filter *bf );

/**
 * @brief Gets the given element's index within internal bit array. 
 * @param bf the Bloom_Filter
 * @param payload the element for which the index is wanted
 * @return the index of the element (crc64 of the element modulo the size of the array)
 */
static inline uint64_t Bloom_Filter_get_cell(  struct Bloom_Filter *bf , void *payload )
{
    uint64_t key = MALP_Trace_crc64( (char *)payload, bf->sizeof_payload );
    
    return (key % bf->size);
}

/**
 * @brief Adds an element to the filter. 
 * @param bf the filter where to add the element
 * @param payload the element to add
 */
static inline void Bloom_Filter_push( struct Bloom_Filter *bf , void *payload )
{
    uint64_t cell = Bloom_Filter_get_cell( bf , payload );
    
    MALP_Spinlock_lock( &bf->lock );
    Bit_Array_set( &bf->array, cell, 1 );
    MALP_Spinlock_unlock( &bf->lock );
}

/**
 * @brief Tests if the filter contains an element. 
 * @param bf the filter chere to search
 * @param payload the element to find in the filter
 * @return 0 if the element is not in the filter, 1 if it may be. 
 */
static inline int Bloom_Filter_test( struct Bloom_Filter *bf , void *payload )
{
    uint64_t ret = 0;
    uint64_t cell = Bloom_Filter_get_cell( bf , payload );

    MALP_Spinlock_lock( &bf->lock );
    ret = Bit_Array_get( &bf->array, cell);
    MALP_Spinlock_unlock( &bf->lock );

    return ret;
}

/**
 * @brief Copies a Bloom_Filter within an other. 
 * @param dest the destination filter
 * @param src the source filter
 *
 * @warning As only the internal bit array is copied, src and dest must be initialized with the same parameters before replicate. 
 */
void Bloom_Filter_replicate( struct Bloom_Filter *dest , struct Bloom_Filter *src );


#ifdef __cplusplus
}
#endif

#endif /* BLOOM_FILTER_H */ 

/**
 * @}
 */
