/* ############################ MALP License ############################## */
/* # Fri Jan 18 14:00:00 CET 2013                                         # */
/* # Copyright or (C) or Copr. Commissariat a l'Energie Atomique          # */
/* #                                                                      # */
/* # This software is governed by the CeCILL-C license under French law   # */
/* # and abiding by the rules of distribution of free software.  You can  # */
/* # use, modify and/ or redistribute the software under the terms of     # */
/* # the CeCILL-C license as circulated by CEA, CNRS and INRIA at the     # */
/* # following URL http://www.cecill.info.                                # */
/* #                                                                      # */
/* # The fact that you are presently reading this means that you have     # */
/* # had knowledge of the CeCILL-C license and that you accept its        # */
/* # terms.                                                               # */
/* #                                                                      # */
/* # Authors:                                                             # */
/* #   - BESNARD Jean-Baptiste jean-baptiste.besnard@cea.fr               # */
/* #                                                                      # */
/* ######################################################################## */
#ifndef BUFFERED_LIFO_H
#define BUFFERED_LIFO_H

/**
 * @addtogroup Buffered_LIFO_
 * @{
 */


/**
 * @addtogroup internal_LIFO_
 * @{
 */

#ifdef __cplusplus
  extern "C"
  {
#endif


#include <stdint.h>
#include <stdlib.h>
#include "Spinlock.h"

/**
 * @brief Structure defining a chunk of LIFO. 
 * A buffered lifo consists in several chunks which are stacked
 * in a lifo fashion. Moreover those chunks gather several elems
 * in order to avoid alocation and dealocation plus it reduces
 * the pointer versus data ratio for small amount of data.
 */
struct Buffered_LIFO_chunk
{
	uint64_t offset; 					/**< @brief offset in current chunk */
	struct Buffered_LIFO_chunk *next; /**< @brief A pointer to the next chunk (NULL if none) */
	char elems[0];						/**< @brief A dummy pointer to the data which is piggybacked */
};

/**
 * @brief Allocates an empty Buffered_LIFO_chunk pointing to prev
 * @param sizeof_payload size of a payload element to compute piggy backing stored in the LIFO to gain space
 * @param chunk_size the number of elements the chunk can contain
 * @param prev pointer to the previous Buffered_LIFO_chunk in order to chain them
 * @return A pointer to the newly allocated chunk
 */
struct Buffered_LIFO_chunk * Buffered_LIFO_chunk_new( size_t sizeof_payload, uint64_t chunk_size, struct Buffered_LIFO_chunk * prev );

/**
 * @brief Releases a Buffered_LIFO_chunk note that it does not recurses over previous chunks
 * @param chunk Chunk to be released
 * @param sizeof_payload size of a payload element 
 * @param free_func a function to be called over each elem still present in the FIFO (NULL if node)
 */
void Buffered_LIFO_chunk_release( struct Buffered_LIFO_chunk * chunk, size_t sizeof_payload, void (*free_func)( void *) );

/**
 * @brief Push a new elem in the Buffered_LIFO_chunk
 * @param chunk Chunk to which we want to add an elem (can be NULL and then leads to a chunk allocation)
 * @param sizeof_payload size of a payload element 
 * @param chunk_size the number of elements the chunk can contain
 * @return A pointer to current chunk (can change if a new one had been allocated)
 */
struct Buffered_LIFO_chunk *  Buffered_LIFO_chunk_push( struct Buffered_LIFO_chunk *  chunk,
														  void * elem,  size_t sizeof_payload, uint64_t chunk_size );

/**
 * @brief Get an element from the Buffered_LIFO_chunk
 * @param chunk Chunk to which we want to add an elem (can be NULL and then leads to did_ret == 0 )
 * @param sizeof_payload size of a payload element 
 * @param chunk_size the number of elements the chunk can contain
 * @param did_ret boolean which allows us to know if we actually retrieved an elem from the LIFO
 * @return A pointer to current chunk (can change if a chunk had been freed)
 */
struct Buffered_LIFO_chunk * Buffered_LIFO_chunk_pop(  struct Buffered_LIFO_chunk * chunk, 
														 size_t sizeof_payload, uint64_t chunk_size,
														 void * elem, int *did_ret );
/**
 * @brief Get a pointer to the Buffered_LIFO_chunk head eleme (NULL if none)
 * @param chunk Chunk from which we want to get the head
 * @return A pointer to current chunk head (NULL if none)
 */
void * Buffered_LIFO_chunk_head( struct Buffered_LIFO_chunk * chunk,  size_t sizeof_payload );

/**
 * @}
 */


/**
 * @brief Structure defining buffered LIFO
 * 
 * Its usage is thread safe and basically wraps a Buffered_LIFO_chunk
 * 
 */
struct Buffered_LIFO
{
	size_t sizeof_payload;					/**< @brief Sizeof a payload element */
	uint64_t chunk_size;					/**< @brief Maximum number of elems in a chunk */
	
	MALP_Spinlock lock;						/**< @brief A spinlock to protect the LIFO */
	
	struct Buffered_LIFO_chunk * head;	/**< @brief A pointer to the first chunk */
};

/**
 * @brief Initializes a Buffered_LIFO ready to contains elems of size sizeof_payload with chunk_size chunks.
 * @param sizeof_payload size of a payload element 
 * @param chunk_size the number of elements the chunk can contain
 */
void Buffered_LIFO_init( struct Buffered_LIFO *lifo, size_t sizeof_payload, uint64_t chunk_size );

/**
 * @brief Releases a Buffered_LIFO and all its chunks
 * @param sizeof_payload size of a payload element 
 * @param chunk_size the number of elements the chunk can contain
 * @param free_func a function to be called over each elem still present in the FIFO (NULL if node)
 */
void Buffered_LIFO_release( struct Buffered_LIFO *lifo, void (*free_func)( void *) );

/**
 * @brief Push an element of size sizeof_payload in the Buffered_LIFO
 * @param lifo pointer to an initialized Buffered_LIFO 
 * @param elem pointer to an elem of sizeof_payload
 */
void Buffered_LIFO_push( struct Buffered_LIFO *lifo, void * elem );

/**
 * @brief Push an element of size sizeof_payload in the Buffered_LIFO
 * @param lifo pointer to an initialized Buffered_LIFO 
 * @param elem pointer to an elem of sizeof_payload to which we want data to be copied
 * @return elem if an element was popped, NULL otherwise
 */
void * Buffered_LIFO_pop( struct Buffered_LIFO *lifo, void * elem );

/**
 * @brief Get a reference to the head element of the Buffered_LIFO
 *        note that as no lock is held this reference can be
 *        invalidated by another thread. In such context you can
 *        use  Buffered_LIFO_head_locked instead.
 * @param lifo pointer to an initialized Buffered_LIFO 
 * @return a pointer to the head element of the LIFO, NULL if none
 */
void * Buffered_LIFO_head( struct Buffered_LIFO *lifo );

/**
 * @brief Get a reference to the head element of the Buffered_LIFO
 *        but keeps the LIFO locked in order to allow thread safe
 *        manipulation of the head. Note that Buffered_LIFO_head_unlock
 *        must be called to unlock the LIFO after !
 * @param lifo pointer to an initialized Buffered_LIFO 
 * @return a pointer to the head element of the LIFO, NULL if none
 */
void * Buffered_LIFO_head_locked( struct Buffered_LIFO *lifo );

/**
 * @brief Unlocks the Buffered_LIFO after a call to Buffered_LIFO_head
 * @param lifo pointer to an initialized Buffered_LIFO 
 */
void Buffered_LIFO_head_unlock( struct Buffered_LIFO *lifo );

#ifdef __cplusplus
	}
#endif

#endif /* BUFFERED_LIFO_H */

/**
 * @}
 */
