/* ############################ MALP License ############################## */
/* # Fri Jan 18 14:00:00 CET 2013                                         # */
/* # Copyright or (C) or Copr. Commissariat a l'Energie Atomique          # */
/* #                                                                      # */
/* # This software is governed by the CeCILL-C license under French law   # */
/* # and abiding by the rules of distribution of free software.  You can  # */
/* # use, modify and/ or redistribute the software under the terms of     # */
/* # the CeCILL-C license as circulated by CEA, CNRS and INRIA at the     # */
/* # following URL http://www.cecill.info.                                # */
/* #                                                                      # */
/* # The fact that you are presently reading this means that you have     # */
/* # had knowledge of the CeCILL-C license and that you accept its        # */
/* # terms.                                                               # */
/* #                                                                      # */
/* # Authors:                                                             # */
/* #   - BESNARD Jean-Baptiste jean-baptiste.besnard@cea.fr               # */
/* #                                                                      # */
/* ######################################################################## */
#define _GNU_SOURCE
#include <dlfcn.h>
#include <unistd.h>
#include <mpi.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>

#include "Timer.h"



#ifndef SYNC_TIME
#define SYNC_TIME

/**
 * @addtogroup Sync_Time_
 *
 * @{
 */

/**
 * @brief the maximum number computations tries
 */
#define MAX_LOOP_TRY 20

/**
 * @brief The tag for time sync communications
 */
#define SYNC_TAG 1234

/**
 * @brief the number of offset computations to perform to get an average offset
 */
#define MAX_AVG 10


/**
 * @brief This structure describes a tree node for time synchronization. 
 */
struct sync_tree_conf {
	int source_node;        /**< @brief the source node */
	int target_node;        /**< @brief the target node */
	int current_pow;        /**< @brief the current nearest power of two */
	int comm_size;          /**< @brief the total number of ranks */
	long long int offset;   /**< @brief the time offset */
};




/**
 * @brief Entry point to be used to perform time sync of every MPI Processes
 */
void sync_mpi();



/**
 * @brief This function does the actual work prepared by bootstrap_sync_tree(). 
 * @param conf the sync_tree_conf corresponding to current node
 * @param action a function to be called on each node
 * @param per_child_act a function to be called on each child of the node
 * @param post_action a function to be called at the end
 * @param arg an argument to give to the three above functions
 * 
 * The work is : 
 *   - run action with conf and arg
 *   - compute a sync_tree_conf struct for each child and send it. (retrieved in bootstrap_sync_tree())
 *   - for each child, run per_child_act with newly computed sync_tree_conf and arg
 *   - run post_action with conf and arg
 *
 * @todo [DOC] describe the tree...
 */
void compute_sync_tree( struct sync_tree_conf *conf,
                        void (*action)(struct sync_tree_conf *next_conf, void *arg),
                        void (*per_child_act)(struct sync_tree_conf *conf, void *arg ),
                        void (*post_action)(struct sync_tree_conf *next_conf, void *arg),
                        void *arg
                      );



/**
 * @brief This is a bootstrapping function for calling compute_sync_tree();
 * @param action a function to be called on each node
 * @param per_child_act a function to be called on each child of the node
 * @param post_action a function to be called at the end
 * @param arg an argument to give to the three above functions
 *
 *   - For rank 0, it performs initialization of a sync_tree_conf struct, then call compute_sync_tree() 
 *   - For other ranks, recieves a sync_tree_conf struct from MPI, and then call compute_sync_tree()
 * 
 * The communucation is performed in a tree manner and rank 0 is the tree root. 
 *
 */
void bootstrap_sync_tree( void (*action)(struct sync_tree_conf *next_conf, void *arg),
                          void (*per_child_act)(struct sync_tree_conf *conf , void *arg),
                          void (*post_action)(struct sync_tree_conf *next_conf, void *arg),
                          void *arg
                        );

//TREE ACTIONS
/**
 * @brief What to be done on each node's child
 * @param c
 * @param arg
 *
 * calls sync_server();
 */
void perchild_host(struct sync_tree_conf *c, void *arg);

/**
 * @brief What to be done on the client side
 * @param c
 * @param arg
 *
 * calls sync_client();
 */
void post_client(struct sync_tree_conf *c, void *arg);

/**
 * @brief Copies offset of c to current process timer offset (poffset)
 * @param c a node configuration
 * @param poffset the current process timer offset
 */
void action_set_off(struct sync_tree_conf *c, void *poffset);
// ~~~ DBG

/**
 * @brief For debugging only
 * @param c a node configuration
 * @param arg an argument
 *
 * Prints who synced with who
 */
void action_val(struct sync_tree_conf *c, void *arg);



//SYNC
/**
 * @brief Server function for synchronizing time
 * @param dest_rank the rank to be synchronized with self
 */
void sync_server( int dest_rank );

/**
 * @brief The client function for synchronizing time (computes offset between server and client)
 * @param parent the parent (server)
 * @param poffset the computed offset
 */
void sync_client(int parent, void *poffset);


/**
 * @cond
 */
extern long long int Process_Sync_Offset;
/**
 * @endcond
 */



//UTILITIES FUNCTIONS

/**
 * @brief Computes an absolute difference
 * @param a first element
 * @param b second element
 * @return the positive difference between a and b
 */
static inline uint64_t abs_diff( uint64_t a , uint64_t b )
{
	if( a < b )
		return b - a;
	else
		return a - b;
}

/**
 * @brief Computes the nearest 
 * @param num the value we want the nearest power of two
 * @return the nearest power of two that is superior to num
 */
static inline uint32_t nearest_pow (uint32_t num)
{
	uint32_t n = num > 0 ? num - 1 : 0;

	n |= n >> 1;
	n |= n >> 2;
	n |= n >> 4;
	n |= n >> 8;
	n |= n >> 16;
	n++;

	return n;
}



/**
 * @}
 */


#endif
