/* ############################ MALP License ############################## */
/* # Fri Jan 18 14:00:00 CET 2013                                         # */
/* # Copyright or (C) or Copr. Commissariat a l'Energie Atomique          # */
/* #                                                                      # */
/* # This software is governed by the CeCILL-C license under French law   # */
/* # and abiding by the rules of distribution of free software.  You can  # */
/* # use, modify and/ or redistribute the software under the terms of     # */
/* # the CeCILL-C license as circulated by CEA, CNRS and INRIA at the     # */
/* # following URL http://www.cecill.info.                                # */
/* #                                                                      # */
/* # The fact that you are presently reading this means that you have     # */
/* # had knowledge of the CeCILL-C license and that you accept its        # */
/* # terms.                                                               # */
/* #                                                                      # */
/* # Authors:                                                             # */
/* #   - BESNARD Jean-Baptiste jean-baptiste.besnard@cea.fr               # */
/* #                                                                      # */
/* ######################################################################## */

#ifndef KNOWLEDGE_SYSTEM_H
#define KNOWLEDGE_SYSTEM_H

#ifdef __cplusplus
extern "C" {
#endif

#include "Buffered_FIFO.h"
#include "Spinlock.h"

struct Data_entry;
struct MALP_blackboard;

/**
 * @ingroup Knowledge_System_
 * @brief The Knowledge_system structure
 *
 * @todo [DOC] What is sensivity ?
 */
struct Knowledge_system
{
    uint8_t freed;                      /**< KS freeing flag */
    int64_t type;                       /**< Knowledge system type (key) */
    MALP_Spinlock lock;                 /**< The Spinlock for concurrent access */

	/* Sensivity Handling */
    uint32_t sensivity_count;           /**< Number of data entries (@ref Data_Entry_ ) */
    uint64_t *sensivity_keys;           /**< Keys of the target data-types */
    uint64_t *sensivity_array;          /**< Array used to detect multiple sensivities  */
    struct Buffered_FIFO *data_entries; /**< Data entries Fifo */
    int is_serial;						/**< Boolean which determines whether jobs have to be processed serially (relativelly to this KS) */

    void (*process)( struct MALP_blackboard *bb, 
					 struct Data_entry **de, uint32_t entry_count, void *arg ); /**< The function to be called on data entries */
    void *arg;                          /**<  arguments to give to process */
};


/**
 * @ingroup Knowledge_System_
 * @brief Initializes a Knowledge_system
 * @param ks the Knowledge_system to initialize
 * @param type the type of the new Knowledge_system (arbitrary integer)
 * @param sensivity_keys the sensivity keys
 * @param sensivity_count the number of data entries (can't be 0)
 * @param process the function to give to the newly created Knowledge_system
 * @param arg the fourth argument to give to process
 */
void Knowledge_system_init( struct Knowledge_system *ks, uint64_t type, uint64_t *sensivity_keys, uint32_t sensivity_count,
			    void (*process)( struct MALP_blackboard *bb, struct Data_entry **de, uint32_t entry_count, void *arg ), void *arg );

/**
 * @ingroup Knowledge_System_
 * @brief Knowledge_system release
 * @param ks
 */
void Knowledge_system_release( struct Knowledge_system *ks );

/**
 * @ingroup Knowledge_System_
 * @brief sets the processing function to a Knowledge_system
 * @param ks the Knowledge_system where to set the function
 * @param process the process to set
 * @param arg an aditionnal parameter to give to process
 */
void Knowledge_system_set_process( struct Knowledge_system *ks, void (*process)( struct MALP_blackboard *bb, struct Data_entry **de, uint32_t entry_count, void *arg ), void *arg );

/**
 * @ingroup Knowledge_System_
 * @brief sets the parallelism mode for this KS
 * @param ks the Knowledge_system to targer
 * @param is_serial whether processing is serial or not (boolean) default YES
 */
void Knowledge_system_set_serial( struct Knowledge_system *ks, int is_serial );

/**
 * @ingroup Knowledge_System_
 * @brief Knowledge_system_has_unsatisfied_sensivities
 * @param ks
 * @return
 */
int Knowledge_system_has_unsatisfied_sensivities( struct Knowledge_system *ks );



/**
 * @ingroup Knowledge_System_
 * @brief struct representing a job that does  the actual data processing (created by Knowledge_system)
 */
struct Knowledge_system_job
{
    uint64_t ks_type;               /**< @brief the type of Knowledge_system */
    int32_t entries_count;          /**< the number of data entries to process */
    struct Data_entry **entries;    /**< the data entries to process */
    void (*process)( struct MALP_blackboard *bb, struct Data_entry **de, uint32_t entry_count, void *arg ); /**< the function that does the work on data entries */
    void *arg;                      /**< the additionnal argument do give to process */
    int is_serial;					/**< Boolean which determines whether jobs have to be processed serially (relativelly to a single KS) */
};

/**
 * @ingroup Knowledge_System_
 * @brief Knowledge_system_job_new
 * @param ks
 * @return
 */
struct Knowledge_system_job Knowledge_system_job_new( struct Knowledge_system *ks );

/**
 * @ingroup Knowledge_System_
 * @brief Knowledge_system_job_release
 * @param ksj
 */
void Knowledge_system_job_release( struct Knowledge_system_job *ksj );


/**
 * @ingroup Knowledge_System_
 * @brief Knowledge_system_submit_data_entry
 * @param ks
 * @param de
 * @return
 */
struct Knowledge_system_job Knowledge_system_submit_data_entry( struct Knowledge_system *ks, struct Data_entry *de );


#ifdef __cplusplus
}
#endif


#endif
