/* ############################ MALP License ############################## */
/* # Fri Jan 18 14:00:00 CET 2013                                         # */
/* # Copyright or (C) or Copr. Commissariat a l'Energie Atomique          # */
/* #                                                                      # */
/* # This software is governed by the CeCILL-C license under French law   # */
/* # and abiding by the rules of distribution of free software.  You can  # */
/* # use, modify and/ or redistribute the software under the terms of     # */
/* # the CeCILL-C license as circulated by CEA, CNRS and INRIA at the     # */
/* # following URL http://www.cecill.info.                                # */
/* #                                                                      # */
/* # The fact that you are presently reading this means that you have     # */
/* # had knowledge of the CeCILL-C license and that you accept its        # */
/* # terms.                                                               # */
/* #                                                                      # */
/* # Authors:                                                             # */
/* #   - BESNARD Jean-Baptiste jean-baptiste.besnard@cea.fr               # */
/* #                                                                      # */
/* ######################################################################## */
#ifndef MALP_BB_H
#define MALP_BB_H

#ifdef __cplusplus
extern "C" {
#endif

#include <stddef.h>
#include <stdint.h>
#include <unistd.h>

#include "Hash_Table.h"
#include "Chained_List.h"
#include "Spinlock.h"

#include "Knowledge_System.h"
#include "Data_Entry.h"
#include "Key_Entry.h"
#include "Pbb_Workers.h"

#include "MALP_Blackboard_Level.h"

/**
 * @ingroup MALP_Blackboard_
 * @brief Number of KS cells (Number of elements)
 */
#define KS_COLL_AVOID 5

/**
 * @ingroup MALP_Blackboard_
 * @brief The Knowledge system cell structure
 * Structure containing an internal hash table and table of knowledge systems ( @ref Knowledge_System_ )
 */
struct Ks_cell
{
    struct Hash_Table_Internal ___;             /**< Internal hash table */
    struct Knowledge_system ks[KS_COLL_AVOID];  /**< Knowledge systems table */
};

/**
 * @ingroup MALP_Blackboard_
 * @brief Number of column keys (Number of elements)
 */
#define KEY_COLL_AVOID 5
/**
 * @ingroup MALP_Blackboard_
 * @brief The Key_cell structure
 * Structure containing an internal hash table and table of key entries ( @ref Key_Entry_ )
 */
struct Key_cell
{
    struct Hash_Table_Internal ___;         /**< Internal hash table */
    struct Key_entry ke[KEY_COLL_AVOID];    /**< Key entries table */
};

/**
 * @ingroup MALP_Blackboard_
 * @brief Number of column names (Number of elements)
 */
#define NAME_COLL_AVOID 5
/**
 * @ingroup MALP_Blackboard_
 * @brief The Name_cell structure
 * Structure containing an internal hash table and table of cell names
 */
struct Name_cell
{
    struct Hash_Table_Internal ___; /**< Internal hash table */
    void *name[NAME_COLL_AVOID];    /**< Name table */
};

/**
 * @ingroup MALP_Blackboard_
 * @brief The MALP_blackboard structure
 */
struct MALP_blackboard
{
    struct Hash_Table ks_table;         /**< Knowledge systems hash table */
    struct Hash_Table key_table;        /**< Keys hash table */
    struct Hash_Table de_name_table;    /**< Descriptions/Names hash table */

    struct Pbb_workers workers;         /**< Blackboard workers */
	
    void (*default_handler)( struct MALP_blackboard *, uint64_t KS_id, struct Data_entry **, uint32_t, void *arg); /**< what to be done by default (if the Knowledge_system_job does not have a handler) */

    void *context;                      /**< Pointer for the current context: @ref Analyzer_BB or Dart_Engine_ (NULL otherwise) */

    volatile uint64_t pending_datas;    /**< Pending datas left flag */
    uint32_t cpu_count;                 /**< Number of currently online processors */

    MALP_Spinlock lock;                 /**< Spin lock */
};


/**
 * @ingroup MALP_Blackboard_
 * @brief MALP_blackboard initialisation
 * @param bb
 * @param workers_count
 * @param context
 */
void MALP_blackboard_init( struct MALP_blackboard *bb, uint32_t workers_count, void *context );

/**
 * @ingroup MALP_Blackboard_
 * @brief sets MALP_blackboard's default @ref Knowledge_System_ handler
 * @param bb the blackboard where to set the default handler
 * @param default_handler the new default handler
 */
static inline void MALP_blackboard_set_default_KS_handler( struct MALP_blackboard *bb, void (*default_handler)( struct MALP_blackboard *, uint64_t KS_id, struct Data_entry **, uint32_t, void * ) )
{
    bb->default_handler = default_handler;
}


/**
 * @ingroup MALP_Blackboard_
 * @brief MALP_blackboard_wait
 * Function to call the wait function for the @ref Pbb_Workers_ of a @ref MALP_Blackboard_
 * @param bb
 */
void MALP_blackboard_wait( struct MALP_blackboard *bb );

/**
 * @ingroup MALP_Blackboard_
 * @brief MALP_blackboard release
 * @param bb
 */
void MALP_blackboard_release( struct MALP_blackboard *bb );


/**
 * @ingroup MALP_Blackboard_
 * @brief Return the number of currently online processors for a MALP_blackboard
 * @param bb
 * @return Number of currently online processors
 */
static inline uint32_t MALP_blackboard_cpu_count(struct MALP_blackboard *bb)
{
	return bb->cpu_count;
}


/**
 * @ingroup MALP_Blackboard_
 * @brief test if a knowledge system is available in a MALP_Blackboard
 * Test if a knowledge system of a specific type, defined by its key, is available in a MALP_Blackboard
 * @param bb Blackboard
 * @param type Knowledge system key
 * @return 1: Knowledge system type defined in a MALP_Blackboard, 0: Knowledge system type not defined
 */
int MALP_Blackboard_ks_present( struct MALP_blackboard *bb, uint64_t type );

/**
 * @ingroup MALP_Blackboard_
 * @brief Create a new @ref Knowledge_System_ while associating a key (used as an id and a type) and a handler (process)
 * @param bb Blackboard
 * @param type Knowledge system key
 * @param sensivities list of sensivities (see @ref Data_Entry_ )
 * @param sensivities_count number of sensivities in the list
 * @param process Handler of a knowledge system
 * @param arg List/handler for arguments
 */
struct Knowledge_system *  MALP_Blackboard_new_ks( struct MALP_blackboard *bb, uint64_t type, uint64_t *sensivities,
							 uint32_t sensivities_count, void (*process)( struct MALP_blackboard *bb, struct Data_entry **de, uint32_t entry_count, void *arg ), void *arg );

/**
 * @ingroup MALP_Blackboard_
 * @brief MALP_Blackboard_delete_ks
 * @param bb
 * @param type
 */
void MALP_Blackboard_delete_ks( struct MALP_blackboard *bb, uint64_t type );

/**
 * @ingroup MALP_Blackboard_
 * @brief MALP_Blackboard_submit_data
 * @param bb
 * @param key
 * @param payload
 * @param payload_size
 * @return
 */
int MALP_Blackboard_submit_data( struct MALP_blackboard *bb, uint64_t key, void *payload, size_t payload_size );

/**
 * @ingroup MALP_Blackboard_
 * @brief MALP_Blackboard_submit_data_entry
 * @param bb
 * @param de
 * @return
 */
int MALP_Blackboard_submit_data_entry(  struct MALP_blackboard *bb, struct Data_entry *de );

/**
 * @brief MALP_Blackboard_resubmit_data_entry
 * @param bb
 * @param de
 * @return
 */
int MALP_Blackboard_resubmit_data_entry(  struct MALP_blackboard *bb, struct Data_entry *de );


/**
 * @ingroup MALP_Blackboard_
 * @brief MALP_Blackboard_writable_data
 * @param de
 * @return
 */
static inline int MALP_Blackboard_writable_data( struct Data_entry *de )
{

	if( Data_entry_get_ref_count( de ) == 0 )
	{
		return 1;
	}
	else
	{
		return 0;
	}
}

/**
 * @ingroup MALP_Blackboard_
 * @brief Returns the number of pebnding data block counter
 * @param bb a pointer to the blacboard
 * @return Number of pending blocks in the BB
 * @return
 */
static inline uint64_t MALP_Blackboard_pending( struct MALP_blackboard *bb )
{
    uint64_t ret = 0;

	MALP_Spinlock_lock( &bb->lock );
	ret = bb->pending_datas;
	MALP_Spinlock_unlock( &bb->lock );

    return ret;
}


/**
 * @ingroup MALP_Blackboard_
 * @brief MALP_Blackboard_wait_pending
 * @param bb
 * @param value
 */
static inline void MALP_Blackboard_wait_pending( struct MALP_blackboard *bb, uint64_t value )
{
	int got_val = 0;
	
	while( !got_val )
	{
		if( MALP_Blackboard_pending( bb ) == 0 )
			got_val = 1;
		
		usleep( 1024 );
	}
}


/**
 * @ingroup MALP_Blackboard_
 * @brief MALP_Blackboard_type_hash
 * @param type
 * @return
 */
static inline uint64_t MALP_Blackboard_type_hash(uint64_t type)
{
       return MALP_Trace_crc64( (char *)&type, sizeof( uint64_t ) );
}
 
/**
 * @ingroup MALP_Blackboard_
 * @brief MALP_Blackboard_type_state_hash
 * @param type
 * @param state
 * @return
 */
static inline uint64_t MALP_Blackboard_type_state_hash(uint64_t type, uint64_t state)
{
       uint64_t a[2] = {type, state};
       return MALP_Trace_crc64( (char *)a, sizeof( uint64_t ) * 2 );
}


/**
 * @ingroup MALP_Blackboard_
 * @brief Increments the number of pending data block counter
 * @param bb a pointer to the blacboard
 * @return
 */
static inline void MALP_Blackboard_add_pending( struct MALP_blackboard *bb)
{
	MALP_Spinlock_lock( &bb->lock );
	bb->pending_datas++;
	MALP_Spinlock_unlock( &bb->lock );
}


/**
 * @ingroup MALP_Blackboard_
 * @brief Decrements the number of pending data block counter
 * @param bb a pointer to the blacboard
 * @return
 */
static inline void MALP_Blackboard_remove_pending( struct MALP_blackboard *bb)
{
	MALP_Spinlock_lock( &bb->lock );
	bb->pending_datas--;
	MALP_Spinlock_unlock( &bb->lock );
}



/* Export test funcs for tests */
/**
 * @ingroup MALP_Blackboard_
 * @brief knowledge_system_test_func
 * @param pks
 * @param key
 * @param arg
 * @return
 */
int knowledge_system_test_func(void *pks, uint64_t key, void *arg);

/**
 * @ingroup MALP_Blackboard_
 * @brief empty_knowledge_system_test_func
 * @param pks
 * @param key
 * @param arg
 * @return
 */
int empty_knowledge_system_test_func(void *pks, uint64_t key, void *arg);

/**
 * @ingroup MALP_Blackboard_
 * @brief key_entry_test_func
 * @param pke
 * @param key
 * @param arg
 * @return
 */
int key_entry_test_func(void *pke, uint64_t key, void *arg);

/**
 * @ingroup MALP_Blackboard_
 * @brief empty_key_entry_test_func
 * @param pke
 * @param key
 * @param arg
 * @return
 */
int empty_key_entry_test_func(void *pke, uint64_t key, void *arg);


#ifdef __cplusplus
}
#endif


#endif
