#include "Red_Dispatch_Bootstrap.h"

#include "Socket_Utils.h"
#include "Red_Tree.h"


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/socket.h>
#include <stdint.h>

char * Red_Tree_Bootstrap_from_env(struct Red_Tree * rt)
{
	char * dispatch_uri = getenv("RT_ADDR");
	
	if( !dispatch_uri )
	{
		printf("Please start the dispatch server and set RT_ADDR before launching server\n");
		abort();
	}
	
	/* Parse uri */
	struct Connection_ctx dispatch_ctx;
	
	if( Connection_ctx_init( &dispatch_ctx, dispatch_uri ) )
	{
		printf("Error parsing RT_ADDR\n");
		abort();
	}
	
	
	
	/* Start a client socket 
	 be very patient as its a centralized server ...*/
	int cnt = 0;
	
	int client_socket = -1;
	
	while( cnt < 1500 )
	{
		client_socket = Socket_connect_to_server( dispatch_ctx.addr,dispatch_ctx.port);
		
		if( 0 < client_socket )
		{
			break;
		}
		
		
		/* Set an increasing timeframe
		 from 10 ms to 1.5 sec*/
		int timeval = cnt * 1000;
		timeval = (timeval < 10000)?10000:timeval;
		
		usleep(timeval);
		cnt+=100;
	}
	
	/* Check that we are connected */
	
	if( client_socket < 0 )
	{
		printf("Error could not connect to %s\n", dispatch_uri );
		abort();
	}
	
	/* Now we can start the status exchange */
	
	/* Send our infos */
	if( Socket_safe_write( client_socket, &rt->listening_socket_ctx, sizeof( struct Connection_ctx ) ) < 0 )
	{
		/* Something went wrong */
		close( client_socket );
		abort();
	}
	
	/* Receive remote infos */
	struct Connection_ctx remote_ctx;
	Connection_ctx_release( &remote_ctx );
	
	if( Socket_safe_read( client_socket, &remote_ctx, sizeof( struct Connection_ctx ) ) < 0 )
	{
		/* Something went wrong */
		close( client_socket );
		abort();
	}
	
		
	/* Whatever happens set instance id */
	rt->instance_id = remote_ctx.id;
	printf("This is id %lld\n", (long long int)rt->instance_id );
	
	/* Now we check if we have a target host */
	if( strlen( remote_ctx.addr ) )
	{
		char target_host[256];
		Connection_ctx_print( &remote_ctx, target_host, 256 );
		return strdup(target_host);
	}
	
	return NULL;
}


void * _Red_Dispatcher_server_loop( void *prd )
{
	struct Red_Dispatcher *rd =  (struct Red_Dispatcher *)prd;
	
	uint64_t counter = 0;
	
	/* Enter acceptance loop */
	while(1)
	{
		int inbound_fd = Socket_wait_for_a_client(rd->listening_socket);
		
		if( inbound_fd < 0 )
		{
			/* Connection closed */
			break;
		}
		
		struct Connection_ctx remote_ctx;
		
		if( Socket_safe_read( inbound_fd, &remote_ctx, sizeof( struct Connection_ctx ) ) < 0 )
		{
			/* Something went wrong */
			close( inbound_fd );
			continue;
		}
		
		printf("New client on %s:%s\n", remote_ctx.addr, remote_ctx.port);
		
		
		struct Node_list registered_node;
		
		/* Create an empty CTx */
		struct Connection_ctx candidate_ctx;
		Connection_ctx_release( &candidate_ctx );
		
		/* Set an unique id before sending back */
		candidate_ctx.id = counter++;
		
		if( Node_list_pop( &rd->nl, &registered_node ) )
		{
			/* There is a candidate in the tree
			 * so we copy back infos */
			printf("DID POP %s:%d\n", registered_node.addr, registered_node.port );
			Connection_ctx_init_static( &candidate_ctx, registered_node.addr, registered_node.port );
		}
		
		/* We now send the Ctx , empty or not */
		if( Socket_safe_write( inbound_fd, &candidate_ctx, sizeof( struct Connection_ctx ) ) < 0 )
		{
			/* Something went wrong */
			close( inbound_fd );
			continue;
		}
		
		printf("Sent back %s:%s\n", candidate_ctx.addr, candidate_ctx.port);
		
		/* Register the new process */
		rd->nl = Node_list_push( rd->nl, remote_ctx.addr, atoi(remote_ctx.port) );
	}
	
	pthread_exit(NULL);
}



int Red_Dispatcher_init( struct Red_Dispatcher *rd )
{
	
	Node_list_init( &rd->nl );
	
	/* Retrieve server CTX */
	
	if( gethostname(rd->server_ctx.addr, 256) < 0 )
	{
		perror("gethostname");
		return 1;
	}
	
	int port = 0;
	
	/* Start server socket */
	rd->listening_socket = Socket_listen_on_free_port(&port);
	
	if( rd->listening_socket < 0 )
	{
		printf("Could not set a listening socket on %s\n", rd->server_ctx.addr);
		return 1;
	}
	
	snprintf(rd->server_ctx.port, 64, "%d", port );
	
	return 0;
}

int Red_Dispatcher_thread(struct Red_Dispatcher* rd)
{
	rd->has_thread = 1;
	
	/* Start server thread */
	if( pthread_create(&rd->server_thread, NULL, _Red_Dispatcher_server_loop, (void *)rd) != 0)
	{
		perror("pthread_create");
		return 1;
	}

	
	return 0;
}


int Red_Dispatcher_loop(struct Red_Dispatcher* rd)
{
	_Red_Dispatcher_server_loop((void *)rd);
	
	return 0;
}



int Red_Dispatcher_release( struct Red_Dispatcher *rd )
{
	shutdown( rd->listening_socket, SHUT_RDWR );
	close( rd->listening_socket );
	
	if( rd->has_thread )
		pthread_join(rd->server_thread, NULL);
	
	Node_list_release(&rd->nl);

	return 0;
}





