#include "Socket_Utils.h"

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>



int Socket_connect_to_server( char* addr, char * port)
{
	int client_socket = -1;
	errno = 0;

	struct addrinfo *results;

	/* First use getaddrinfo to extract connection type */
	int ret = getaddrinfo( addr, port, NULL, &results);

	if( ret != 0 )
	{
		printf("Error resolving name : %s\n", gai_strerror(ret) );
		return -1;
	}

	struct addrinfo *current = results;

	int connected = 0;


	/* And now try to connect to the server in a portable way */
	while( current )
	{
		errno = 0;
		/* First open a socket */
		client_socket = socket( current->ai_family, current->ai_socktype, current->ai_protocol );

		printf("Trying %d %d %d\n", current->ai_family, current->ai_socktype, current->ai_protocol );
		
		if( client_socket < 0 )
		{
			perror( "socket" );
		}
		else
		{
			/* And now lets try to connect */
			errno = 0;
			if (connect(client_socket, current->ai_addr, current->ai_addrlen) != -1)
			{
				printf("Connected\n");
				connected = 1;
				/* We are connected all ok */
				break;
			}
			else
			{
				perror("connect");
			}

			/* If we are here something went wront */
			close( client_socket );
		}

		/* Lets try next connection type */
		current = current->ai_next;
	}

	freeaddrinfo(results);

	if( connected == 0 )
	{
		printf("Failed to connect to %s:%lld\n", addr, (long long int)port);
		return -1;
	}
	
	return client_socket;
}

int Socket_listen_on_free_port(int *port)
{
	int listening_socket = 0;
	/* Reset errno to 0 */
	errno = 0;

	if( !port )
	{
		printf("Error port pointer is NULL\n");
		abort();
	}
	
	char s_port[64];
	snprintf(s_port, 64, "%d", *port);
	
	/* Use a portable way of setting a listening port */
	struct addrinfo hints, *results;

	/* Initialize the addrinfo hints struct to what we expect */
	memset( &hints, 0, sizeof( struct addrinfo ) );
	hints.ai_family = AF_UNSPEC; /* Whatever type of network is available */
	hints.ai_socktype = SOCK_STREAM; /* We want to use it as a TCP socket */
	hints.ai_flags = AI_PASSIVE; /* We want to bind this socket */

	/* Note that we bind ourselves to port 0 in order to get a free port */
	int ret = getaddrinfo( NULL, s_port, &hints, &results);

	if( ret != 0 )
	{
		printf("Could not find a binding method : %s\n", gai_strerror(ret) );
		return -1;
	}

	struct addrinfo *current = results;
	
	/* We will now iterate over results to start listening on the first reliable candidate */
	while( current )
	{
		errno = 0;
		/* First try to open a socket */
		listening_socket = socket( current->ai_family, current->ai_socktype, current->ai_protocol );

		if( listening_socket < 0 )
		{
			perror( "socket" );
		}
		else
		{

			errno = 0;
			
			/* Bind the socket to the port */
			if (bind(listening_socket, current->ai_addr, current->ai_addrlen) != -1)
			{
					/* Start listening on that port */
					if (listen(listening_socket,  512) != -1)
					{
							/* As we are bound to "0" we want to get the actual port */
							struct sockaddr_in socket_infos;
							socklen_t infolen = sizeof(struct sockaddr);
							
							if(getsockname(listening_socket, (struct sockaddr*)&socket_infos, &infolen) == -1 )
							{
								perror( "getsockname" );
								return -1;
							}
							else
							{
								if( infolen == sizeof(struct sockaddr_in) )
								{
									*port = ntohs(socket_infos.sin_port);
								}
								else
								{
									printf("Could not retrieve port number\n");
									close(listening_socket);
									return -1;
								}
							}   

							break;
			
					}
					else
					{
						perror("listen");
						close(listening_socket);
						
					}
			}
			else
			{
				perror("bind");
				close(listening_socket);
			}
		}
		
		current = current->ai_next;
	}

	freeaddrinfo(results);

	if( listening_socket == 0 )
	{
		printf("Failed to set a listening socket\n");
		return -1;
	}

	return listening_socket;
}

int Socket_wait_for_a_client(int listening_socket)
{
	socklen_t len = sizeof( struct sockaddr );
	struct sockaddr remote_infos;

	int client_socket =  accept(listening_socket, &remote_infos, &len);

	if( client_socket == -1 )
	{
		return -1;
	}
	
	/* Here a client is connected */
	return client_socket;
}



int Socket_safe_read( int socket, void *dest, size_t size )
{
	int ret = recv( socket, dest, size, MSG_WAITALL );
	
	if( ret < 0 )
	{
		perror("recv");
		return -1;
	}

	return ret;
}

int Socket_safe_write( int socket, void *src, size_t size )
{
	int ret = send( socket, src, size, 0 );
	
	if( ret < 0 )
	{
		perror("send");
		return -1;
	}

	return ret;
}

