/* ############################ MALP License ############################## */
/* # Fri Jan 18 14:00:00 CET 2013                                         # */
/* # Copyright or (C) or Copr. Commissariat a l'Energie Atomique          # */
/* #                                                                      # */
/* # This software is governed by the CeCILL-C license under French law   # */
/* # and abiding by the rules of distribution of free software.  You can  # */
/* # use, modify and/ or redistribute the software under the terms of     # */
/* # the CeCILL-C license as circulated by CEA, CNRS and INRIA at the     # */
/* # following URL http://www.cecill.info.                                # */
/* #                                                                      # */
/* # The fact that you are presently reading this means that you have     # */
/* # had knowledge of the CeCILL-C license and that you accept its        # */
/* # terms.                                                               # */
/* #                                                                      # */
/* # Authors:                                                             # */
/* #   - BESNARD Jean-Baptiste jean-baptiste.besnard@cea.fr               # */
/* #                                                                      # */
/* ######################################################################## */



#ifndef VMPI_STREAMS
#define VMPI_STREAMS

#include <mpi.h>

#ifdef __cplusplus
extern "C" {
#endif

#include <stdlib.h>
#include <VMPI.h>
#include <Spinlock.h>

/**
 * @ingroup VMPI_Streams_
 * @brief Defines the load ballancing modes available to MPI_Streams
 */
typedef enum
{
	VMPI_STREAM_BALANCE_NONE,			/**< @brief Always send or receive from the same */
	VMPI_STREAM_BALANCE_RANDOM,			/**< @brief Send or receive from a random one */
	VMPI_STREAM_BALANCE_ROUND_ROBIN		/**< @brief Send or receive from each turn by turn */
}VMPI_Stream_load_balance;

/**
 * @ingroup VMPI_Streams_
 * @brief Defines MPI_Stream mode
 */
typedef enum
{
	VMPI_STREAM_R,	/**< @brief READ */
	VMPI_STREAM_W,	/**< @brief WRITE */
	VMPI_STREAM_RW	/**< @brief READ and WRITE */
}VMPI_Stream_mode;


/**
 * @ingroup VMPI_Streams_
 * @brief Defines call mode
 */
typedef enum
{
	VMPI_STREAM_BLOCK,		/**< @brief All Calls are blocking ones */
	VMPI_STREAM_NONBLOCK	/**< @brief If no data are available or endpoint not ready
							   call might return VMPI_EAGAIN */
}VMPI_Stream_blocking;

/**
 * @ingroup VMPI_Streams_
 * @brief Internal declaration of a stream endpoint
 */
struct _VMPI_Stream
{
	int target;		/**< @brief Target task */
	int remote_tag;	/**< @brief Remote unique tag */
	int local_tag;	/**< @brief Local unique tag */
	int state;		/**< @brief Been closed or not */
};

/**
 * @ingroup VMPI_Streams_
 * @brief Maximum number of asynchronous cells
 */
#define VMPI_STREAM_MAX_ASYNC 10

/**
 * @ingroup VMPI_Streams_
 * @brief Defines a MPI stream
 *
 * Note that there is VMPI_STREAM_MAX_ASYNC buffers per input to avoid
 * unexpected messages whereas only VMPI_STREAM_MAX_ASYNC buffers
 * are available for output and are therefore shared between multiple endpoints
 *
 */
typedef struct
{
	struct _VMPI_Stream *in;					/**< @brief Input streams */
	int current_in;								/**< @brief Current input stream */
	int in_count;								/**< @brief Number of inputs */

	struct _VMPI_Stream *out;					/**< @brief Output streams */
	int current_out;							/**< @brief Current output stream */
	int out_count;								/**< @brief Number of outputs */

	size_t block_size;							/**< @brief Size of a data block */

	char **r_tmp_buffers;						/**< @brief Temporary buffer for input */
	int *r_current_buff;						/**< @brief Current buffer */
	MPI_Request *r_req;							/**< @brief Input requests */
	MPI_Status *r_stat;							/**< @brief Input statuses */
	int r_req_count;							/**< @brief Number of input requests */


	char *w_tmp_buffers[VMPI_STREAM_MAX_ASYNC];	/**< @brief Output buffers */
	MPI_Request w_req[VMPI_STREAM_MAX_ASYNC];	/**< @brief Output requests */
	MPI_Status w_stat[VMPI_STREAM_MAX_ASYNC];	/**< @brief Output statuses */
	int w_req_count;							/**< @brief Number of output requests */

	VMPI_Stream_load_balance lb;				/**< @brief Load balancing policy */

    MALP_Spinlock lock;                         /**< @brief Spinlock */
}VMPI_Stream;

/**
 * @ingroup VMPI_Streams_Manipulations_
 * @brief Initializes a stream should be called before any *open call
 * @param st target stream
 * @param block_size size of a data block
 * @param lb load balancing policy
 * @return 0 on success
 */
int VMPI_Stream_init( VMPI_Stream *st, size_t block_size, VMPI_Stream_load_balance lb);

/**
 * @ingroup VMPI_Streams_Manipulations_
 * @brief Open a stream to dest rank, dest should do the symmetrical call
 * @param st target stream
 * @param dest target rank
 * @param mode read, write or both
 * @return 0 on success
 */
int VMPI_Stream_open( VMPI_Stream *st, int dest, char *mode );

/**
 * @ingroup VMPI_Streams_Manipulations_
 * @brief Open a stream according to a mapping
 * @param st target stream
 * @param map mapping to refer to
 * @param mode read, write or both
 * @return 0 on success
 */
int VMPI_Stream_open_map( VMPI_Stream *st, VMPI_Map *map, char *mode );

/**
 * @ingroup VMPI_Streams_Manipulations_
 * @brief Closes a stream and sends EOF to remote endpoints
 * @param st target stream
 * @return 0 on success, 1 if PMPI_Send fails, 2 if malloc fails
 */
int VMPI_Stream_close( VMPI_Stream * st );


/**
 * @ingroup VMPI_Streams_Manipulations_
 * @brief Join two streams in the same set
 * @param dest the destination stream (where src will be joined)
 * @param src the source stream (the stream to be joined into dest)
 * @return ?
 *
 * @warning Afert this call the src handler shall not be used anymore
 * @todo Clean release of src
 */
int VMPI_Stream_join(VMPI_Stream * dest, VMPI_Stream * src);

/**
 * @ingroup VMPI_Streams_Manipulations_
 * @fn int VMPI_Stream_push_stream( VMPI_Stream *st, int dest, int remote_tag, int local_tag, VMPI_Stream_mode mode )
 * @brief pushes a stream
 * @param st the stream to push
 * @param dest the task where to push the stream
 * @param remote_tag the tag of the remote task
 * @param local_tag the local tag
 * @param mode the mode of the stream
 * @return 0 on success, 2 on allocation error
 * @todo [DOC] Explainations on VMPI_Stream_push_stream...
 */

/**
 * @ingroup VMPI_Read_Write_
 * @brief Read data from a VMPI_Stream
 * @param st target stream
 * @param buff buffer to write to
 * @param block_count number of blocks
 * @param block should this call be a blocking one
 * @return the size of read data if positive, a @link VMPI_Return_Status @endlink otherwise
 */
int VMPI_Stream_read( VMPI_Stream * st, char *buff, unsigned int block_count, VMPI_Stream_blocking block );

/**
 * @ingroup VMPI_Read_Write_
 * @brief Write data to a VMPI_Stream
 * @param st target stream
 * @param buff buffer to read from
 * @param block_count number of blocks
 * @return the size of written data, -1 if fails
 */
int VMPI_Stream_write( VMPI_Stream * st, char *buff, unsigned int block_count );

/**
 * @ingroup VMPI_Read_Write_
 * @brief Is this stream ready for read ?
 * @param st target stream
 * @return 1 if the stream is ready for reading, 0 if not
 */
int VMPI_Stream_test_read( VMPI_Stream * st );

/**
 * @ingroup VMPI_Read_Write_
 * @brief Is this stream ready for writing ?
 * @param st target stream
 * @return 1 if the stream is ready for writing, 0 if not
 */
int VMPI_Stream_test_write( VMPI_Stream * st );

/**
 * @ingroup VMPI_Getters_
 * @brief Get the number of input streams
 * @param st target stream
 * @return the number of input streams
 */
static inline int VMPI_Stream_read_count( VMPI_Stream * st )
{
	return st->in_count;
}

/**
 * @ingroup VMPI_Getters_
 * @brief Get the number of output streams
 * @param st target stream
 * @return the number of output streams
 */
static inline int VMPI_Stream_write_count( VMPI_Stream * st )
{
	return st->out_count;
}

#ifdef __cplusplus
}
#endif

#endif /* VMPI_STREAMS */

